from visidata import VisiData, vd

vd.theme_option('disp_splitwin_pct', 0, 'height of second sheet on screen')
vd.theme_option('disp_note_none', '⌀',  'visible contents of a cell whose value is None')
vd.theme_option('disp_truncator', '…', 'indicator that the contents are only partially visible')
vd.theme_option('disp_oddspace', '\u00b7', 'displayable character for odd whitespace')
vd.theme_option('disp_more_left', '<', 'header note indicating more columns to the left')
vd.theme_option('disp_more_right', '>', 'header note indicating more columns to the right')
vd.theme_option('disp_error_val', '', 'displayed contents for computation exception')
vd.theme_option('disp_ambig_width', 1, 'width to use for unicode chars marked ambiguous')

vd.theme_option('disp_pending', '', 'string to display in pending cells')
vd.theme_option('disp_note_pending', ':', 'note to display for pending cells')
vd.theme_option('disp_note_fmtexc', '?', 'cell note for an exception during formatting')
vd.theme_option('disp_note_getexc', '!', 'cell note for an exception during computation')
vd.theme_option('disp_note_typeexc', '!', 'cell note for an exception during type conversion')

vd.theme_option('color_note_pending', 'bold green', 'color of note in pending cells')
vd.theme_option('color_note_type', '226 yellow', 'color of cell note for non-str types in anytype columns')
vd.theme_option('color_note_row', '220 yellow', 'color of row note on left edge')
vd.option('scroll_incr', -3, 'amount to scroll with scrollwheel')
vd.theme_option('disp_column_sep', '│', 'separator between columns')
vd.theme_option('disp_keycol_sep', '║', 'separator between key columns and rest of columns')
vd.theme_option('disp_rowtop_sep', '│', '') # ╷│┬╽⌜⌐▇
vd.theme_option('disp_rowmid_sep', '⁝', '') # ┃┊│█
vd.theme_option('disp_rowbot_sep', '⁝', '') # ┊┴╿⌞█⍿╵⎢┴⌊  ⋮⁝
vd.theme_option('disp_rowend_sep', '║', '') # ┊┴╿⌞█⍿╵⎢┴⌊
vd.theme_option('disp_keytop_sep', '║', '') # ╽╿┃╖╟
vd.theme_option('disp_keymid_sep', '║', '') # ╽╿┃
vd.theme_option('disp_keybot_sep', '║', '') # ╽╿┃╜‖
vd.theme_option('disp_endtop_sep', '║', '') # ╽╿┃╖╢
vd.theme_option('disp_endmid_sep', '║', '') # ╽╿┃
vd.theme_option('disp_endbot_sep', '║', '') # ╽╿┃╜‖
vd.theme_option('disp_selected_note', '•', '') #
vd.theme_option('disp_sort_asc', '↑↟⇞⇡⇧⇑', 'characters for ascending sort') # ↑▲↟↥↾↿⇞⇡⇧⇈⤉⤒⥔⥘⥜⥠⍏˄ˆ
vd.theme_option('disp_sort_desc', '↓↡⇟⇣⇩⇓', 'characters for descending sort') # ↓▼↡↧⇂⇃⇟⇣⇩⇊⤈⤓⥕⥙⥝⥡⍖˅ˇ
vd.theme_option('color_default', 'white on black', 'the default fg and bg colors')
vd.theme_option('color_default_hdr', 'bold white on black', 'color of the column headers')
vd.theme_option('color_bottom_hdr', 'underline white on black', 'color of the bottom header row')
vd.theme_option('color_current_row', 'reverse', 'color of the cursor row')
vd.theme_option('color_current_col', 'bold on 232', 'color of the cursor column')
vd.theme_option('color_current_cell', '', 'color of current cell, if different from color_current_row+color_current_col')
vd.theme_option('color_current_hdr', 'bold reverse', 'color of the header for the cursor column')
vd.theme_option('color_column_sep', 'white on black', 'color of column separators')
vd.theme_option('color_key_col', '81 cyan', 'color of key columns')
vd.theme_option('color_hidden_col', '8', 'color of hidden columns on metasheets')
vd.theme_option('color_selected_row', '215 yellow', 'color of selected rows')
vd.theme_option('color_clickable', 'bold', 'color of internally clickable item')
vd.theme_option('color_code', 'bold white on 237', 'color of code sample')
vd.theme_option('color_heading', 'bold black on yellow', 'color of header')
vd.theme_option('color_guide_unwritten', '243 on black', 'color of unwritten guides in GuideGuide')

vd.theme_option('force_256_colors', False, 'use 256 colors even if curses reports fewer')

vd.option('quitguard', False, 'confirm before quitting modified sheet')
vd.option('default_width', 20, 'default column width', replay=True)   # TODO: make not replay and remove from markdown saver
vd.option('default_height', 4, 'default column height')



'Enter edit mode automatically when typing numeric digits.'

from visidata import Sheet

for i in range(0, 10):
    Sheet.addCommand(str(i), 'autoedit-%s' % i, 'cursorCol.setValues([cursorRow], editCell(cursorVisibleColIndex, value="%s", i=1))' % i, 'replace cell value with input starting with %s' % i)



import itertools
import copy
import datetime
import re
from colorsys import rgb_to_hls, hls_to_rgb

from visidata import VisiData, vd, Sheet, Column, Progress, IndexSheet, ColumnAttr, SequenceSheet, AttrDict, AttrColumn
from visidata import CellColorizer, getattrdeep, rgb_to_attr
from visidata.type_date import date


vd.option('xlsx_meta_columns', False, 'include columns for cell objects, font colors, and fill colors', replay=True)
vd.option('xlsx_color_cells', True, 'color cells based on xlsx source')

@VisiData.api
def open_xls(vd, p):
    p.is_local() or vd.fail('xls loader does not support remote files')
    return XlsIndexSheet(p.base_stem, source=p)

@VisiData.api
def open_xlsx(vd, p):
    p.is_local() or vd.fail('xlsx loader does not support remote files')
    return XlsxIndexSheet(p.base_stem, source=p)

class XlsxIndexSheet(IndexSheet):
    'Load XLSX file (in Excel Open XML format).'
    rowtype = 'sheets'  # rowdef: xlsxSheet
    columns = [
        Column('sheet', getter=lambda col,row: row.source.title),  # xlsx sheet title
        ColumnAttr('name', width=0),  # visidata Sheet name
        ColumnAttr('nRows', type=int),
        ColumnAttr('nCols', type=int),
        Column('active', getter=lambda col,row: row.source is col.sheet.workbook.active),
    ]
    nKeys = 1

    def iterload(self):
        openpyxl = vd.importExternal('openpyxl')
        self.workbook = openpyxl.load_workbook(str(self.source), data_only=True, read_only=True)
        for sheetname in self.workbook.sheetnames:
            src = self.workbook[sheetname]
            yield XlsxSheet(self.name, sheetname, source=src, workbook=self.workbook)


class XlsxSheet(SequenceSheet):
    # rowdef: AttrDict of column_letter to cell
    colorizers = [
        CellColorizer(5, None, lambda s,c,r,v: c and r and s.colorize_xlsx_cell(c,r))
    ]
    def setCols(self, headerrows):
        vd.importExternal('openpyxl')
        from openpyxl.utils.cell import get_column_letter
        self.columns = []
        self._rowtype = AttrDict

        if not headerrows:
            return

        headers = [[cell.value for cell in row.values()] for row in headerrows]
        column_letters = [
                x.column_letter if 'column_letter' in dir(x)
                else get_column_letter(i+1)
                for i, x in enumerate(headerrows[0].values())]

        for i, colnamelines in enumerate(itertools.zip_longest(*headers, fillvalue='')):
            colnamelines = ['' if c is None else c for c in colnamelines]
            column_name = ''.join(map(str, colnamelines))
            self.addColumn(AttrColumn(column_name, column_letters[i] + '.value', column_letter=column_letters[i]))
            self.addXlsxMetaColumns(column_letters[i], column_name)

    def addRow(self, row, index=None):
        Sheet.addRow(self, row, index=index)  # skip SequenceSheet
        for column_letter, v in list(row.items())[len(self.columns):len(row)]:  # no-op if already done
            self.addColumn(AttrColumn('', column_letter + '.value'))
            self.addXlsxMetaColumns(column_letter, column_letter)

    def iterload(self):
        vd.importExternal('openpyxl')
        from openpyxl.utils.cell import get_column_letter
        worksheet = self.source
        for row in Progress(worksheet.iter_rows(), total=worksheet.max_row or 0):
            yield AttrDict({get_column_letter(i+1): cell for i, cell in enumerate(row)})

    def addXlsxMetaColumns(self, column_letter, column_name):
        if self.options.xlsx_meta_columns:
            self.addColumn(
                    AttrColumn(column_name + '_cellPyObj', column_letter, column_letter=column_letter))
            self.addColumn(
                    AttrColumn(column_name + '_fontColor',
                        column_letter + '.font.color.value', column_letter=column_letter))
            self.addColumn(
                    AttrColumn(column_name + '_fillColor', column_letter +
                        '.fill.start_color.value', column_letter=column_letter))
            self.addColumn(Column(column_name + '_colorizer', width=0,
                           column_letter=column_letter,
                           getter=lambda c,r: c.sheet.colorize_xlsx_cell(c,r)))

    def paste_after(self, rowidx):
        to_paste = list(copy.copy(r) for r in reversed(vd.memory.cliprows))
        self.addRows(to_paste, index=rowidx)


class XlsIndexSheet(IndexSheet):
    'Load XLS file (in Excel format).'
    rowtype = 'sheets'  # rowdef: xlsSheet
    columns = [
        Column('sheet', getter=lambda col,row: row.source.name),  # xls sheet name
        ColumnAttr('name', width=0),  # visidata sheet name
        ColumnAttr('nRows', type=int),
        ColumnAttr('nCols', type=int),
    ]
    nKeys = 1
    def iterload(self):
        xlrd = vd.importExternal('xlrd')
        self.workbook = xlrd.open_workbook(str(self.source))
        for sheetname in self.workbook.sheet_names():
            yield XlsSheet(self.name, sheetname, source=self.workbook.sheet_by_name(sheetname))


class XlsSheet(SequenceSheet):
    def iterload(self):
        worksheet = self.source
        for rownum in Progress(range(worksheet.nrows)):
            yield list(worksheet.cell(rownum, colnum).value for colnum in range(worksheet.ncols))


@Sheet.property
def xls_name(vs):
    name = vs.names[-1]
    if vs.options.clean_names:
        cleaned_name = ''.join('_' if ch in ':[]*?/\\' else ch for ch in vs.name) #1122
        name = cleaned_name[:31] #1122  #594
        name = name.strip('_')

    return name


@VisiData.api
def save_xlsx(vd, p, *sheets):
    openpyxl = vd.importExternal('openpyxl')

    wb = openpyxl.Workbook()
    wb.remove_sheet(wb['Sheet'])

    def _convert_save_row(dispvals:dict, replace_illegal=False) -> list:
        row = []
        for col, v in dispvals.items():
            if v is None:
                v = ""
            elif col.type == date:
                v = datetime.datetime.fromtimestamp(int(v.timestamp()))
            elif not vd.isNumeric(col):
                v = str(v)
                if replace_illegal:
                    v = re.sub(openpyxl.cell.cell.ILLEGAL_CHARACTERS_RE, ' ', v)

            row.append(v)
        return row

    for vs in sheets:
        if vs.xls_name != vs.names[-1]:
            vd.warning(f'saving {vs.name} as {vs.xls_name}')
        ws = wb.create_sheet(title=vs.xls_name)

        headers = [col.name for col in vs.visibleCols]
        ws.append(headers)

        for dispvals in vs.iterdispvals(format=False):
            row = _convert_save_row(dispvals)
            try:
                ws.append(row)
            except openpyxl.utils.exceptions.IllegalCharacterError as e:
                row = _convert_save_row(dispvals, replace_illegal=True)  #1402
                ws.append(row)

    wb.active = ws

    wb.save(filename=p)


@VisiData.api
def save_xls(vd, p, *sheets):
    xlwt = vd.importExternal('xlwt')

    wb = xlwt.Workbook()

    for vs in sheets:
        if vs.xls_name != vs.name:
            vd.warning(f'saving {vs.name} as {vs.xls_name}')
        ws1 = wb.add_sheet(vs.xls_name)
        for col_i, col in enumerate(vs.visibleCols):
            ws1.write(0, col_i, col.name)

        for r_i, dispvals in enumerate(vs.iterdispvals(format=True)):
            r_i += 1
            for c_i, v in enumerate(dispvals.values()):
                ws1.write(r_i, c_i, v)

    wb.save(p)


# from https://stackoverflow.com/a/65426130

RGBMAX = 255
HLSMAX = 240

@XlsxSheet.api
def colorize_xlsx_cell(sheet, col, row):
    if not hasattr(col, 'column_letter') or not sheet.options.xlsx_color_cells:
        return ''
    fg = getattrdeep(row, col.column_letter+'.font.color', None)
    bg = getattrdeep(row, col.column_letter+'.fill.start_color', None)
    fg = sheet.xlsx_color_to_xterm256(fg)
    bg = sheet.xlsx_color_to_xterm256(bg)

    if bg == '-1' or fg == '-1':
        fg, bg = '-1', '-1'

    return f'{fg} on {bg}'

@XlsxSheet.api
def xlsx_color_to_xterm256(sheet, color) -> str:
    if not color:
        return ''

    if color.type == 'rgb':
        s = color.value
        if isinstance(s, int):
            return str(s)

        a,r,g,b = s[0:2], s[2:4], s[4:6], s[6:8]
        return rgb_to_attr(int(r, 16), int(g, 16), int(b, 16), int(a, 16))

    if color.type == 'theme':
        return sheet.theme_and_tint_to_rgb(color.value, color.tint)
    else:
        return str(color.value)

@XlsxSheet.api
def theme_and_tint_to_rgb(sheet, theme, tint) -> str:
    """Given a workbook, a theme number and a tint return a xterm256 color number"""
    rgb = sheet.theme_colors[theme]
    h, l, s = rgb_to_ms_hls(rgb)
    r, g, b = ms_hls_to_rgb(h, tint_luminance(tint, l), s)

    return rgb_to_attr(r*256, g*256, b*256)

@XlsxSheet.lazy_property
def theme_colors(sheet):
    """Gets theme colors from the workbook"""
    # see: https://groups.google.com/forum/#!topic/openpyxl-users/I0k3TfqNLrc
    from openpyxl.xml.functions import QName, fromstring
    xlmns = 'http://schemas.openxmlformats.org/drawingml/2006/main'
    root = fromstring(sheet.workbook.loaded_theme)
    themeEl = root.find(QName(xlmns, 'themeElements').text)
    colorSchemes = themeEl.findall(QName(xlmns, 'clrScheme').text)
    firstColorScheme = colorSchemes[0]

    theme_colors = []

    for c in ['lt1', 'dk1', 'lt2', 'dk2', 'accent1', 'accent2', 'accent3', 'accent4', 'accent5', 'accent6']:
        accent = firstColorScheme.find(QName(xlmns, c).text)
        for i in list(accent): # walk all child nodes, rather than assuming [0]
            if 'window' in i.attrib['val']:
                theme_colors.append(i.attrib['lastClr'])
            else:
                theme_colors.append(i.attrib['val'])

    return theme_colors

def rgb_to_ms_hls(red, green=None, blue=None):
    """Converts rgb values in range (0,1) or a hex string of the form '[#aa]rrggbb' to HLSMAX based HLS, (alpha values are ignored)"""
    if green is None:
        if isinstance(red, str):
            if len(red) > 6:
                red = red[-6:]  # Ignore preceding '#' and alpha values
            blue = int(red[4:], 16) / RGBMAX
            green = int(red[2:4], 16) / RGBMAX
            red = int(red[0:2], 16) / RGBMAX
        else:
            red, green, blue = red
    h, l, s = rgb_to_hls(red, green, blue)
    return (int(round(h * HLSMAX)), int(round(l * HLSMAX)), int(round(s * HLSMAX)))

def ms_hls_to_rgb(hue, lightness=None, saturation=None):
    """Converts HLSMAX based HLS values to rgb values in the range (0,1)"""
    if lightness is None:
        hue, lightness, saturation = hue
    return hls_to_rgb(hue / HLSMAX, lightness / HLSMAX, saturation / HLSMAX)

def rgb_to_hex(red, green=None, blue=None):
    """Converts (0,1) based RGB values to a hex string 'rrggbb'"""
    if green is None:
        red, green, blue = red
    return ('%02x%02x%02x' % (int(round(red * RGBMAX)), int(round(green * RGBMAX)), int(round(blue * RGBMAX)))).upper()

def tint_luminance(tint, lum):
    """Tints a HLSMAX based luminance"""
    # See: http://ciintelligence.blogspot.co.uk/2012/02/converting-excel-theme-color-and-tint.html
    if tint < 0:
        return int(round(lum * (1.0 + tint)))
    else:
        return int(round(lum * (1.0 - tint) + (HLSMAX - HLSMAX * (1.0 - tint))))



'''
Hello world minimal plugin.  Press F2 to show options.hello_world on the status line.

.visidatarc: `import plugins.hello`
'''
from visidata import vd, BaseSheet

vd.option('hello_world', '¡Hola mundo!', 'shown by the hello-world command')

BaseSheet.addCommand('F2', 'hello-world', 'status(options.hello_world)', 'print greeting to status')



import os
import shutil
import stat
import subprocess
import contextlib
try:
    import pwd
    import grp
except ImportError:
    pass # pwd,grp modules not available on Windows

from visidata import Column, Sheet, LazyComputeRow, asynccache, BaseSheet, vd
from visidata import Path, ENTER, asyncthread, VisiData
from visidata import modtime, filesize, vstat, Progress, TextSheet
from visidata.type_date import date


vd.option('dir_depth', 0, 'folder recursion depth on DirSheet')
vd.option('dir_hidden', False, 'load hidden files on DirSheet')


@VisiData.api
def guess_dir(vd, p):
    if p.is_dir():
        return dict(filetype='dir', _likelihood=10)


@VisiData.lazy_property
def currentDirSheet(p):
    'Support opening the current DirSheet from the vdmenu'
    return DirSheet(Path('.').absolute().name, source=Path('.'))

@asyncthread
def exec_shell(*args):
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if err or out:
        lines = err.decode('utf8').splitlines() + out.decode('utf8').splitlines()
        vd.push(TextSheet(' '.join(args), source=lines))


@VisiData.api
def open_dir(vd, p):
    if p.is_dir():
        return DirSheet(p.base_stem, source=p)
    if p.is_file():
        vd.status(f'opening {p.given} as txt')
        return vd.open_txt(p)
    vd.warning(f'could not determine file type for {p.given}')
    return DirSheet(p.base_stem, source=p)

@VisiData.api
def open_fdir(vd, p):
    return FileListSheet(p.base_stem, source=p)

@VisiData.api
def addShellColumns(vd, cmd, sheet, curcol=None):
    shellcol = ColumnShell(cmd, source=sheet, width=0, curcol=curcol)
    sheet.addColumnAtCursor(
            Column(cmd+'_stdout', type=bytes.rstrip, srccol=shellcol, getter=lambda col,row: col.srccol.getValue(row)[0]),
            Column(cmd+'_stderr', type=bytes.rstrip, srccol=shellcol, getter=lambda col,row: col.srccol.getValue(row)[1]),
            shellcol)


class ColumnShell(Column):
    def __init__(self, name, cmd=None, curcol=None, **kwargs):
        super().__init__(name, **kwargs)
        self.expr = cmd or name
        self.curcol = curcol

    @asynccache(lambda col,row: (col, col.sheet.rowid(row)))
    def calcValue(self, row):
        try:
            import shlex
            args = []
            context = LazyComputeRow(self.source, row, curcol=self.curcol)
            for arg in shlex.split(self.expr):
                if arg.startswith('$'):
                    arg = shlex.quote(str(context[arg[1:]]))
                args.append(arg)

            p = subprocess.Popen([os.getenv('SHELL', 'bash'), '-c', shlex.join(args)],
                    stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return p.communicate()
        except Exception as e:
            vd.exceptionCaught(e)


class DirSheet(Sheet):
    'Sheet displaying directory, using ENTER to open a particular file.  Edited fields are applied to the filesystem.'
    guide = '''
        # Directory Sheet
        This is a list of files in the {sheet.displaySource} folder.

        - {help.commands.open_row_file}
        - {help.commands.open_rows}
        - {help.commands.open_dir_parent}
        - {help.commands.sysopen_row}

        ## Options (must reload to take effect)

        - {help.options.dir_depth}
        - {help.options.dir_hidden}
    '''
    rowtype = 'files' # rowdef: Path
    defer = True
    columns = [
        Column('directory',
            getter=lambda col,row: str(row.parent) if str(row.parent) in ('.', '/') else str(row.parent) + '/',
            setter=lambda col,row,val: col.sheet.moveFile(row, val)),
        Column('filename',
            getter=lambda col,row: row._path.name,
            setter=lambda col,row,val: col.sheet.renameFile(row, val)),
        Column('abspath', width=0, type=str,
            getter=lambda col,row: row,
            setter=lambda col,row,val: os.rename(row, val)),
        Column('ext', getter=lambda col,row: row.is_dir() and '/' or row.ext),
        Column('size', type=int,
            getter=lambda col,row: filesize(row),
            setter=lambda col,row,val: os.truncate(row, int(val))),
        Column('modtime', type=date,
            getter=lambda col,row: modtime(row),
            setter=lambda col,row,val: os.utime(row, times=((row.stat().st_atime, float(val))))),
        Column('owner', width=0,
            getter=lambda col,row: pwd.getpwuid(row.stat().st_uid).pw_name,
            setter=lambda col,row,val: os.chown(row, pwd.getpwnam(val).pw_uid, -1)),
        Column('group', width=0,
            getter=lambda col,row: grp.getgrgid(row.stat().st_gid).gr_name,
            setter=lambda col,row,val: os.chown(row, -1, grp.getgrnam(val).pw_gid)),
        Column('mode', width=0,
            getter=lambda col,row: '{:o}'.format(row.stat().st_mode),
            setter=lambda col,row,val: os.chmod(row, int(val, 8))),
        Column('filetype', width=0, cache='async', getter=lambda col,row: subprocess.Popen(['file', '--brief', row], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0].strip()),
    ]
    nKeys = 2
    _ordering = [('modtime', True), ('filename', False)]  # sort by reverse modtime initially

    @staticmethod
    def colorOwner(sheet, col, row, val):
        ret = ''
        if col.name == 'group':
            mode = row.stat().st_mode
            if mode & stat.S_IXGRP: ret = 'bold '
            if mode & stat.S_IWGRP: return ret + 'green'
            if mode & stat.S_IRGRP: return ret + 'yellow'
        elif col.name == 'owner':
            mode = row.stat().st_mode
            if mode & stat.S_IXUSR: ret = 'bold '
            if mode & stat.S_IWUSR: return ret + 'green'
            if mode & stat.S_IRUSR: return ret + 'yellow'

    def moveFile(self, row, newparent):
        parent = Path(newparent)
        newpath = Path(parent/(row.name + row.suffix))
        if parent.exists():
            if not parent.is_dir():
                vd.error('destination %s not a directory' % parent)
        else:
            with contextlib.suppress(FileExistsError):
                os.makedirs(parent)

        row.rename(newpath)
        row.given = newpath # modify visidata.Path
        self.restat()

    def renameFile(self, row, val):
        newpath = row.with_name(val)
        row.rename(newpath)
        row.given = newpath
        self.restat()

    def removeFile(self, path):
        if path.is_dir():
            if self.options.safety_first:
                os.rmdir(path)
            else:
                shutil.rmtree(path)  #1965
        else:
            path.unlink()

    def commitDeleteRow(self, r):
        self.removeFile(r)

    def newRow(self):
        vd.fail('new file not supported')

    def iterload(self):
        hidden_files = self.options.dir_hidden

        def _walkfiles(p, dir_depth:int=0):
            basepath = str(p)
            for folder, subdirs, files in os.walk(basepath):
                subfolder = folder[len(basepath)+1:]
                if not hidden_files and subfolder.startswith('.'): continue
                if subfolder in ['.', '..']: continue

                fpath = Path(folder)
                if str(fpath) != str(p):
                    yield fpath

                for fn in files:
                    yield fpath/fn

                if dir_depth < len(fpath.parents)-len(p.parents)+1:
                    for d in subdirs:
                        yield fpath/d
                    subdirs.clear()

        basepath = str(self.source)

        folders = set()

        for p in _walkfiles(self.source, self.options.dir_depth):
            if not hidden_files and str(p).startswith('.') and not str(p).startswith('..'):
                continue

            yield p

    def preloadHook(self):
        super().preloadHook()
        Path.stat.cache_clear()

    def restat(self):
        vstat.cache_clear()

    @asyncthread
    def putChanges(self):
        self.commitAdds()
        self.commitMods()
        self.commitDeletes()

        self._deferredDels.clear()
        self.reload()

    def getDefaultSaveName(sheet):
        return sheet.name + '.' + sheet.options.save_filetype


class FileListSheet(DirSheet):
    _ordering = []
    def iterload(self):
        for fn in self.source.open():
            yield Path(fn.rstrip())


@VisiData.api
def inputShell(vd):
    cmd = vd.input("sh$ ", type="sh")
    if '$' not in cmd:
        vd.warning('no $column in command')
    return cmd

DirSheet.addCommand('`', 'open-dir-parent', 'vd.push(openSource(source.parent if source.resolve()!=Path(".").resolve() else os.path.dirname(source.resolve())))', 'open parent directory')  #1801
BaseSheet.addCommand('', 'open-dir-current', 'vd.push(vd.currentDirSheet)', 'open Directory Sheet: browse properties of files in current directory')

Sheet.addCommand('z;', 'addcol-shell', 'cmd=inputShell(); addShellColumns(cmd, sheet, curcol=cursorCol)', 'create new column from bash expression, with $columnNames as variables')

DirSheet.addCommand(ENTER, 'open-row-file', 'vd.push(openSource(cursorRow or fail("no row"), filetype="dir" if cursorRow.is_dir() else LazyComputeRow(sheet, cursorRow).ext))', 'open current file as a new sheet')
DirSheet.addCommand('g'+ENTER, 'open-rows', 'for r in selectedRows: vd.push(openSource(r))', 'open selected files as new sheets')
DirSheet.addCommand('^O', 'sysopen-row', 'launchEditor(cursorRow)', 'open current file in external $EDITOR')
DirSheet.addCommand('g^O', 'sysopen-rows', 'launchEditor(*selectedRows)', 'open selected files in external $EDITOR')

DirSheet.addCommand('y', 'copy-row', 'copy_files([cursorRow], inputPath("copy to dest: "))', 'copy file to given directory *path*')
DirSheet.addCommand('gy', 'copy-selected', 'copy_files(selectedRows, inputPath("copy to dest: ", value=cursorRow.given))', 'copy selected files to given directory *path*')

DirSheet.addCommand('z'+ENTER, 'open-row-filetype', 'ft = input("filetype: ", type="filetype", value=options.filetype or LazyComputeRow(sheet, cursorRow).ext); vd.push(openSource(cursorRow, filetype=ft) or fail(f"file {cursorDisplay} does not exist"))', 'open file in current row as input filetype')


@DirSheet.api
@asyncthread
def copy_files(sheet, paths, dest):
    destdir = Path(dest)
    destdir.is_dir() or vd.fail('target must be directory')
    vd.status('copying %s %s to %s' % (len(paths), sheet.rowtype, destdir))
    os.makedirs(destdir, exist_ok=True)
    for srcpath in Progress(paths, gerund='copying'):
        try:
            destpath = destdir/str(srcpath._path.name)
            if srcpath.is_dir():
                shutil.copytree(srcpath, destpath)
            else:
                shutil.copyfile(srcpath, destpath)
        except Exception as e:
            vd.exceptionCaught(e)


vd.addGlobals({
    'DirSheet': DirSheet
})

vd.addMenuItems('''
    Column > Add column > shell > addcol-shell
    Open > file in row > open-row-filetype
''')



import itertools
import functools

from visidata import Sheet, Column, vd, asyncthread, Progress

@Sheet.api
def window(sheet, before:int=0, after:int=0):
    '''Generate (row, list[row]) for each row in *sheet*, where list[row] is the rows within *before* number of rows before and *after* number of rows after the *row*.  The *row* itself is always included in the list.'''
    for i, r in enumerate(sheet.rows):
        a = max(0, i-before) if before >= 0 else None
        b = (i+after+1) if after >= 0 else None
        yield r, sheet.rows[a:b]


@Column.api
def window(col, before:int=0, after:int=0):
    'Generate (row, list[values]) for each row in the sheet.  Values are the typed values for this column at that row.'
    for r, rows in col.sheet.window(before, after):
        yield r, [col.getTypedValue(x) for x in rows]


class WindowColumn(Column):
    def getValue(self, row):
        return self.windowrows.get(id(row), None)

    @asyncthread
    def _calcWindowRows(self, outvals):
        for row, vals in Progress(self.sourcecol.window(self.before, self.after), total=self.sheet.nRows):
            self.windowrows[id(row)] = vals

    @property
    def windowrows(self):
        if not hasattr(self, '_windowrows'):
            self._windowrows = {}
            self._calcWindowRows(self._windowrows)

        return self._windowrows

    def __getstate__(self):
        r = super().__getstate__()
        r['before'] = self.before
        r['after'] = self.after
        r['sourcecol'] = self.sourcecol.name
        return r

    def __setstate__(self, r):
        self.sourcecol = self.sheet.column(r.pop('sourcecol', None))
        return super().__setstate__(r)

@Sheet.api
def addcol_window(sheet, curcol):
    winsizestr = vd.input('# rows before/after window: ', value='1 1')
    before, after = map(int, winsizestr.split())
    newcol = WindowColumn(curcol.name+"_window", sourcecol=curcol, before=before, after=after)
    sheet.addColumnAtCursor(newcol)


@Sheet.api
def select_around(sheet, n):
    sheet.select(list(itertools.chain(*(winrows for row, winrows in sheet.window(int(n), int(n)) if sheet.isSelected(row)))))


Sheet.addCommand('w', 'addcol-window', 'addcol_window(cursorCol)', 'add column where each row contains a list of that row, nBefore rows, and nAfter rows')
Sheet.addCommand('', 'select-around-n', 'select_around(input("select rows around selected: ", value=1))', 'select additional N rows before/after each selected row')

vd.addMenuItem('Row', 'Select', 'N rows around each selected row', 'select-around-n')

vd.addGlobals(WindowColumn=WindowColumn)



import collections
from visidata import Column, Sheet, VisiData, ColumnItem, Progress, TypedExceptionWrapper, SettableColumn
from visidata import asyncthread, vd


@Column.api
def resetCache(col):
    col._cachedValues = collections.OrderedDict()
    vd.status("reset cache for " + col.name)


@Sheet.api
def freeze_col(sheet, col):
    frozencol = SettableColumn(col.name+'_frozen')
    state = col.__getstate__()
    state.pop('name')
    frozencol.__setstate__(state)
    frozencol.recalc(sheet)

    @asyncthread
    def calcRows_async(frozencol, rows, col):
        # no need to undo, addColumn undo is enough
        for r in Progress(rows, 'calculating'):
            try:
                frozencol.putValue(r, col.getTypedValue(r))
            except Exception as e:
                frozencol.putValue(r, e)

    calcRows_async(frozencol, sheet.rows, col)
    return frozencol


@VisiData.api
class StaticSheet(Sheet):
    'A copy of the source sheet with all cells frozen.'
    def __init__(self, source):
        super().__init__(source.name + "'", source=source)

    def resetCols(self):
        self.columns = []
        for i, col in enumerate(self.source.visibleCols):
            colcopy = ColumnItem(col.name)
            colcopy.__setstate__(col.__getstate__())
            colcopy.expr = i
            self.addColumn(colcopy)
            if col in self.source.keyCols:
                self.setKeys([colcopy])

    def iterload(self):
        for r in Progress(self.source.rows, 'calculating'):
            row = []
            yield row

            # now fill out row
            for col in self.source.visibleCols:
                val = col.getTypedValue(r)
                if isinstance(val, TypedExceptionWrapper):
                    row.append(None)
                else:
                    row.append(val)


@Sheet.api
def setcol_freeze(sheet, unfrozen):  #2660  Contributed by @midichef
    frozen = sheet.freeze_col(unfrozen)
    frozen.name = unfrozen.name
    unfrozen.hide()
    vd.addUndoColNames([unfrozen])
    unfrozen.name = frozen.name + '_unfrozen'
    sheet.addColumnAtCursor(frozen)
    vd.status(f'replaced {frozen.name} with frozen copy')


Sheet.addCommand("z'", 'setcol-freeze', 'setcol_freeze(cursorCol)', 'replace current column with a frozen copy, with all cells evaluated')
Sheet.addCommand("'", 'freeze-col', 'sheet.addColumnAtCursor(freeze_col(cursorCol))', 'add a frozen copy of current column with all cells evaluated')
Sheet.addCommand("g'", 'freeze-sheet', 'vd.push(StaticSheet(sheet)); status("pushed frozen copy of "+name)', 'open a frozen copy of current sheet with all visible columns evaluated')
Sheet.addCommand(None, 'cache-col', 'cursorCol.resetCache()', 'add/reset cache for current column')
Sheet.addCommand("gz'", 'cache-cols', 'for c in visibleCols: c.resetCache()', 'add/reset cache for all visible columns')

vd.addMenuItem('Column', 'Freeze', 'setcol-freeze')
vd.addMenuItem('File', 'Freeze', 'freeze-sheet')



'''
This plugin supports the default log format for:

        /var/log/ltm*
        /var/log/gtm*
        /var/log/apm*
        /var/log/audit*

It extracts common log entries, particularly around monitoring, iRules and configuration change audits. It tries to extract data into common fields to assist rapid filtering.

f5log_object_regex provides a simple way to perform a regex on an object name extracted by a splitter and get extra columns out of it. This is very useful when objectnames have a structure. Simply use named groups in your regex to get named columns out.

Regex: (?:/Common/)(?P<site>[^-]+)-(?P<vstype>[^-]+)-(?P<application>[^-]+)

/Common/newyork-www-banking1

... | site    | vstype | application | ...
... | newyork | www    | banking1   | ...

Adding to .visidatarc

echo 'visidata.vd.options.set("f5log_object_regex", r"(?:/Common/)(?P<site>[^-]+)-(?P<vstype>[^-]+)-(?P<application>[^-]+)", obj="global")' > ~/.visidatarc
'''

__author__ = "James Deucker <me@bitwisecook.org>"
__version__ = "1.0.10"

from datetime import datetime, timedelta
from ipaddress import ip_address

import re
import traceback
from typing import Any, Dict, Optional

from visidata import Path, VisiData, Sheet, date, AttrColumn, vd, Column, CellColorizer, RowColorizer


class hexint(int):
    def __new__(cls, value, *args, **kwargs):
        return super(cls, cls).__new__(cls, value, base=16)

    def __str__(self):
        return hex(self)


class delta_t(int):
    def __new__(cls, value, *args, **kwargs):
        return super(cls, cls).__new__(cls, value, base=10)


vd.addType(hexint, icon="ⓧ", formatter=lambda fmt, num: str(num))
vd.addType(
    delta_t,
    icon="⇥",
    formatter=lambda fmt, delta: str(timedelta(seconds=delta)),
)
vd.theme_option("color_f5log_mon_up", "green", "color of f5log monitor status up")
vd.theme_option("color_f5log_mon_down", "red", "color of f5log monitor status down")
vd.theme_option("color_f5log_mon_unknown", "blue", "color of f5log monitor status unknown")
vd.theme_option("color_f5log_mon_checking", "magenta", "color of monitor status checking")
vd.theme_option("color_f5log_mon_disabled", "black", "color of monitor status disabled")
vd.theme_option("color_f5log_logid_alarm", "red", "color of alarms")
vd.theme_option("color_f5log_logid_warn", "yellow", "color of warnings")
vd.theme_option("color_f5log_logid_notice", "cyan", "color of notice")
vd.theme_option("color_f5log_logid_info", "green", "color of info")
vd.option(
    "f5log_object_regex",
    None,
    "A regex to perform on the object name, useful where object names have a structure to extract. Use the (?P<foo>...) named groups form to get column names.",
    help='regex'
)
vd.option(
    "f5log_log_year",
    None,
    "Override the default year used for log parsing. Use all four digits of the year (e.g., 2022). By default (None) use the year from the ctime of the file, or failing that the current year.",
)
vd.option(
    "f5log_log_timezone",
    "UTC",
    "The timezone the source file is in, by default UTC.",
)


class F5LogSheet(Sheet):
    class F5LogRow:
        def __init__(
            self,
            msg: str = None,
            timestamp: datetime = None,
            host: str = None,
            level: str = None,
            process: str = None,
            proc_pid: int = None,
            logid1: hexint = None,
            logid2: hexint = None,
            message: str = None,
            kv: Optional[Dict[str, Any]] = None,
            **kwargs,
        ):
            self._data = {
                "msg": msg,
                "timestamp": timestamp,
                "host": host,
                "level": level,
                "process": process,
                "proc_pid": proc_pid,
                "logid1": logid1,
                "logid2": logid2,
                "message": message,
                "kv": kv,
                **kwargs,
            }

        def __getattr__(self, item):
            return self._data.get(
                item, self._data["kv"].get(item) if self._data["kv"] else None
            )

    # strptime is slow so we need to parse manually
    _months = {
        "Jan": 1,
        "Feb": 2,
        "Mar": 3,
        "Apr": 4,
        "May": 5,
        "Jun": 6,
        "Jul": 7,
        "Aug": 8,
        "Sep": 9,
        "Oct": 10,
        "Nov": 11,
        "Dec": 12,
    }

    _proto = {
        6: "tcp",
        17: "udp",
        132: "sctp",
    }

    rowtype = "logs"

    columns = [
        AttrColumn("rawmsg", type=str, width=0),
        AttrColumn("timestamp", type=date),
        AttrColumn("host", type=str),
        AttrColumn("level", type=str),
        AttrColumn("process", type=str, width=10),
        AttrColumn("proc_pid", type=int, width=7),
        AttrColumn("logid1", type=hexint),
        AttrColumn("logid2", type=hexint),
        AttrColumn("message", type=str, width=90),
        AttrColumn("object", type=str, width=50),
    ]

    re_f5log = re.compile(
        r"^(?:(?:audit|gtm|ltm|security|tmm|user|\<\d+\>)\s+)?(?:(?P<date1>[A-Z][a-z]{2}\s+\d{1,2}\s+\d{2}:\d{2}:\d{2})|(?P<date2>\d+-\d+-\d+T\d+:\d+:\d+[+-]\d+:\d+)|(?P<date3>\d+-\d+\s\d{2}:\d{2}:\d{2}))\s+(?P<host>\S+)\s+(?:(?P<level>\S+)\s+(?:(?P<process>[a-z0-9_()-]+\s?)\[(?P<pid>\d+)\]:\s+)?(?:(?P<logid1>[0-9a-f]{8}):(?P<logid2>[0-9a-f]):\s+)?)?(?P<message>.*)$"
    )
    re_ltm_irule = re.compile(
        r"(?:(?P<irule_msg>TCL\serror|Rule|Pending\srule):?\s(?P<irule>\S+)\s\<(?P<event>[A-Z_0-9]+)\>(?:\s-\s|:\s|\s)?)(?P<message>aborted\sfor\s(?P<srchost>\S+)\s->\s(?P<dsthost>\S+)|.*)"
    )
    re_ltm_pool_mon_status_msg = re.compile(
        r"^(Pool|Node)\s(?P<poolobj>\S+)\s(member|address)\s(?P<poolmemberobj>\S+)\smonitor\sstatus\s(?P<newstatus>.+)\.\s\[\s(?:(?:(?:(?P<monitorobj>\S+):\s(?P<monitorstatus>\w+)(?:;\slast\serror:\s\S*\s?(?P<lasterr>[^\]]*))?)?(?:,\s)?)+)?\s]\s*(?:\[\swas\s(?P<prevstatus>.+)\sfor\s(?P<durationhr>-?\d+)hrs?:(?P<durationmin>-?\d+)mins?:(?P<durationsec>-?\d+)sec\s\])?$"
    )
    re_ltm_ip_msg = re.compile(
        r"(?:.*?)(?P<ip1>\d+\.\d+\.\d+\.\d+)(?:[:.](?P<port1>\d+))?(?:(?:\s->\s|:)(?P<ip2>\d+\.\d+\.\d+\.\d+)(?:[:.](?P<port2>\d+))?)?(?:\smonitor\sstatus\s(?P<mon_status>\w+)\.\s\[[^]]+\]\s+\[\swas\s(?P<prev_status>\w+)\sfor\s((?P<durationhr>-?\d+)hrs?:(?P<durationmin>-?\d+)mins?:(?P<durationsec>-?\d+)secs?)\s\]|\.?(?:.*))"
    )
    re_ltm_conn_error = re.compile(
        r"^(?:\(null\sconnflow\):\s)?Connection\serror:\s(?P<func>[^:]+):(?P<funcloc>[^:]+):\s(?P<error>.*)\s?\((?P<errno>\d+)\)(?:\s(?P<errormsg>.*))$"
    )
    re_ltm_cert_expiry = re.compile(
        r"Certificate\s'(?P<cert_cn>.*)'\sin\sfile\s(?P<file>\S+)\s(?P<message>will\sexpire|expired)\son\s(?P<date1>[A-Z][a-z]{2}\s+\d{1,2}\s+\d{2}:\d{2}:\d{2}\s\d+\s\S+)"
    )
    re_gtm_monitor = re.compile(
        r"^(?:SNMP_TRAP:\s)?(?P<objtype>VS|Pool|Monitor|Wide\sIP|Server|Data\scenter|Prober\sPool|Box)\s\(?(?P<object>\S+?)\)?\s(?:member\s\(?(?P<pool_member>\S+?)\)?\s)?(?:\(ip(?::port)?=(?P<ipport>[^\)]+)\)\s)?(?:\(Server\s(?P<server>[^\)]+)\)\s)?(?:state\schange\s)?(?:(?P<prev_status>\w+)\s-->\s)?(?P<new_status>\w+)(?:(?:\s\(\s?)(?P<msg>(?:(?P<type>\w+)\s(?P<monitor_object>\S+)\s:\s)?state:\s(?P<state>\S+)|.*)\))?$"
    )
    re_gtm_monitor_instance = re.compile(
        r"^Monitor\sinstance\s(?P<object>\S+)\s(?P<monip>\S+)\s(?P<prevstatus>\S+)\s-->\s(?P<newstatus>\S+)\sfrom\s(?P<srcgtm>\S+)\s\((?:state:?\s)?(?P<state>.*)\)$"
    )
    re_ltm_poolnode_abled = re.compile(
        r"^(?P<objtype>Pool|Node|Monitor)\s(?P<object>\S+)\s(?:address|member|instance)\s(?P<member>\S+)\s(session\sstatus|has\sbeen)\s(?P<status>.+)\.$"
    )
    re_ltm_no_shared_ciphers = re.compile(
        r"^(?P<msg>No\sshared\sciphers\sbetween\sSSL\speers)\s(?P<srchost>\d+\.\d+\.\d+\.\d+|[0-9a-f:]+)\.(?P<srcport>\d+)\:(?P<dsthost>\d+\.\d+\.\d+\.\d+|[0-9a-f:]+)\.(?P<dstport>\d+)\.$"
    )
    re_ltm_http_process_state = re.compile(
        r"^http_process_state_(?P<httpstate>\S+)\s-\sInvalid\saction:0x(?P<actionid>[a-f0-9]+)\s(?P<msg>.*?)\s*(?P<sidea>(?:server|client)side)\s\((?P<src>\S+)\s->\s(?P<vsdst>\S+)\)\s+(?:(?P<sideb>(?:server|client)side)\s)?(?:\((?P<poolsrc>\S+)\s->\s(?P<dst>\S+)\)|\(\(null\sconnflow\)\))\s\((?P<sideaa>\S+)\sside:\svip=(?P<vs>\S+)\sprofile=(?P<profile>\S+)\s(?:pool=(?P<pool>\S+)\s)?(?P<sideaaa>\S+)_ip=(?P<sideasrc>\S+)\)$"
    )
    re_ltm_http_header_exceeded = re.compile(
        r"^(?P<msg>HTTP\sheader\s(?:count|\((?P<size>\d+)\))\sexceeded\smaximum\sallowed\s(?P<type>count|size)\sof\s(?P<limit>\d+))\s\((?P<side>\S+)\sside:\svip=(?P<object>\S+)\sprofile=(?P<profile>\S+)\spool=(?P<pool>\S+)\s(?P<sideip>client|server)_ip=(?P<sidehost>.*)\)$"
    )

    f5log_mon_colors = {
        ("monitor_status", "down"): "color_f5log_mon_down",
        ("monitor_status", "up"): "color_f5log_mon_up",
        ("monitor_status", "enabled"): "color_f5log_mon_up",
        ("monitor_status", "forced disabled"): "color_f5log_mon_disabled",
        ("monitor_status", "node disabled"): "color_f5log_mon_disabled",
        ("monitor_status", "checking"): "color_f5log_mon_checking",
        ("new_status", "available"): "color_f5log_mon_up",
        ("new_status", "unavailable"): "color_f5log_mon_down",
        ("new_status", "up"): "color_f5log_mon_up",
        ("new_status", "down"): "color_f5log_mon_down",
        ("new_status", "green"): "color_f5log_mon_up",
        ("new_status", "red"): "color_f5log_mon_down",
        ("new_status", "now has available members"): "color_f5log_mon_up",
        ("new_status", "no members available"): "color_f5log_mon_down",
        ("new_status", "blue"): "color_f5log_mon_unknown",
        ("new_status", "checking"): "color_f5log_mon_checking",
        ("new_status", "unchecked"): "color_f5log_mon_unknown",
        ("new_status", "node down"): "color_f5log_mon_disabled",
        ("new_status", "forced down"): "color_f5log_mon_disabled",
        ("new_status", "disabled"): "color_f5log_mon_disabled",
        ("prev_status", "available"): "color_f5log_mon_up",
        ("prev_status", "unavailable"): "color_f5log_mon_down",
        ("prev_status", "up"): "color_f5log_mon_up",
        ("prev_status", "down"): "color_f5log_mon_down",
        ("prev_status", "green"): "color_f5log_mon_up",
        ("prev_status", "red"): "color_f5log_mon_down",
        ("prev_status", "now has available members"): "color_f5log_mon_up",
        ("prev_status", "no members available"): "color_f5log_mon_down",
        ("prev_status", "blue"): "color_f5log_mon_unknown",
        ("prev_status", "checking"): "color_f5log_mon_checking",
        ("prev_status", "unchecked"): "color_f5log_mon_unknown",
        ("prev_status", "node down"): "color_f5log_mon_disabled",
        ("prev_status", "forced down"): "color_f5log_mon_disabled",
        ("prev_status", "disabled"): "color_f5log_mon_disabled",
    }

    def colorizeMonitors(sheet, col: Column, row: F5LogRow, value):
        if row is None or col is None:
            return None
        return sheet.f5log_mon_colors.get((col.name, value.value), None)

    f5log_warn_logid = {
        "01010013": "color_f5log_logid_notice",
        "01010029": "color_f5log_logid_warn",
        "01010038": "color_f5log_logid_warn",
        "01010201": "color_f5log_logid_warn",
        "01010281": "color_f5log_logid_warn",
        "01070333": "color_f5log_logid_warn",
        "01070596": "color_f5log_logid_alarm",
        "010c0018": "color_f5log_logid_warn",
        "010c0019": "color_f5log_logid_info",
        "010c003e": "color_f5log_logid_alarm",
        "010c003f": "color_f5log_logid_alarm",
        "010c0044": "color_f5log_logid_warn",
        "010c0052": "color_f5log_logid_warn",
        "010c0053": "color_f5log_logid_info",
        "010c0054": "color_f5log_logid_alarm",
        "010c0055": "color_f5log_logid_alarm",
        "010c0057": "color_f5log_logid_info",
        "010e0001": "color_f5log_logid_alarm",
        "010e0004": "color_f5log_logid_alarm",
        "01340011": "color_f5log_logid_warn",
        "01390002": "color_f5log_logid_notice",
        "01140029": "color_f5log_logid_alarm",
        "01140030": "color_f5log_logid_warn",
        "01140045": "color_f5log_logid_alarm",
        "01190004": "color_f5log_logid_alarm",
        "011ae0f3": "color_f5log_logid_alarm",
        "011e0002": "color_f5log_logid_alarm",
        "011e0003": "color_f5log_logid_alarm",
        "011f0005": "color_f5log_logid_warn",
        "014f0004": "color_f5log_logid_warn",
    }

    def colorizeRows(sheet, col: Column, row: F5LogRow, value):
        if row is None or col is None:
            return None
        if (
            row.logid1 is None
            and row.message is not None
            and row.message.startswith("boot_marker")
        ):
            return "color_f5log_logid_notice"
        return sheet.f5log_warn_logid.get(row.logid1, None)

    @staticmethod
    def split_audit_bigip_tmsh_audit(msg):
        # skip 'AUDIT - ' at the start of the line
        e = msg[8:].split("=", maxsplit=6)

        for ee, ne in zip(e, e[1:]):
            yield {ee[ee.rfind(" ") + 1 :]: ne[: ne.rfind(" ")]}

    @staticmethod
    def split_audit_scriptd_run_script(msg):
        # skip 'AUDIT - ' at the start of the line
        e = msg[8:].split("=")

        for ee, ne in zip(e, e[1:]):
            yield {ee[ee.rfind(" ") + 1 :]: ne[: ne.rfind(" ")].strip('"')}

    @staticmethod
    def split_audit_mcpd_mcp_error(msg):
        # skip 'AUDIT - ' at the start of the line
        # skip the status at the end of the line
        msg = msg[8:]
        status = None
        status_loc = msg.rfind("[Status=")
        if status_loc >= 0:
            status = msg[status_loc + 1 : -1]
            msg = msg[: status_loc - 1]
            yield {
                "status": status.split("=", maxsplit=1)[1],
            }
        # we need to get one word back from the first opening curly
        # find the curly
        cmd_data_loc = msg.find(" { ")
        if cmd_data_loc >= 0:
            # get the cmd_data
            cmd_data = msg[cmd_data_loc + 1 :]
            # split the message and the command
            msg, cmd = msg[:cmd_data_loc].rsplit(" ", maxsplit=1)
            # strip off the trailing " -" from the message
            msg = msg[:-2]
            object = cmd_data.split('"', maxsplit=2)
            if len(object) == 3:
                yield {"object": object[1]}
            yield {
                "command": cmd,
                "cmd_data": cmd_data,
            }

        e = msg.split(" - ")

        for ee in e[0].split(","):
            ee = ee.strip().split(" ")
            # yield the kvs in the first bit split on ,
            if ee[0].startswith("tmsh-pid-"):
                # of course tmsh-pid- is different
                yield {ee[0][: ee[0].rfind("-")]: int(ee[0][ee[0].rfind("-") + 1 :])}
            elif len(ee) == 1:
                yield {ee[0]: None}
            else:
                yield {ee[0]: ee[1]}

        for ee in e[1:]:
            ee = ee.strip().split(" ", maxsplit=1)
            if ee[0] == "transaction":
                yield {"transaction": int(ee[1][1:].split("-")[0])}
                yield {"transaction_step": int(ee[1][1:].split("-")[1])}
            elif ee[0] == "object":
                yield {"object_id": ee[1]}
            else:
                # yield the rest of the kvs
                try:
                    yield {ee[0]: ee[1]}
                except IndexError:
                    yield {ee[0]: None}

    @staticmethod
    def split_ltm_pool_mon_status(msg):
        m = F5LogSheet.re_ltm_pool_mon_status_msg.match(msg)
        if m is None:
            return
        m = m.groupdict()
        if m.get("durationhr") and m.get("durationmin") and m.get("durationsec"):
            duration = timedelta(
                hours=int(m.get("durationhr")),
                minutes=int(m.get("durationmin")),
                seconds=int(m.get("durationsec")),
            ).total_seconds()
        else:
            duration = None
        dst = m.get("poolmemberobj")
        if dst:
            dst = dst.split("/")[-1]
            if "." in dst and len(dst.split(":")) == 2:
                # ipv4
                dsthost, dstport = dst.split(":")
            elif "." in dst and len(dst.split(":")) == 1:
                # ipv4
                dsthost, dstport = dst, None
            else:
                # ipv6
                dsthost, dstport = dst.rsplit(":", maxsplit=1)
            try:
                # see if it's an IP and if so parse it
                dsthost = ip_address(dsthost)
            except ValueError:
                dsthost = None
            try:
                # see if it's a port number and if so parse it
                dstport = int(dstport)
            except (ValueError, TypeError):
                dstport = None
        else:
            dsthost, dstport = None, None
        yield {
            "object": m.get("poolobj"),
            "objtype": "pool",
            "pool_member": m.get("poolmemberobj"),
            "monitor": m.get("monitorobj"),
            "dsthost": dsthost,
            "dstport": dstport,
            "monitor_status": m.get("monitorstatus"),
            "new_status": m.get("newstatus"),
            "prev_status": m.get("prevstatus"),
            "last_error": m.get("lasterr"),
            "duration_s": duration,
        }

    @staticmethod
    def split_ltm_poolnode_mon_abled(msg):
        m = F5LogSheet.re_ltm_poolnode_abled.match(msg)
        if m is None:
            return
        m = m.groupdict()
        yield {
            "object": m.get("object"),
            "objtype": m.get("objtype").lower(),
            "pool_member": m.get("member"),
            "monitor_status": m.get("status"),
        }

    @staticmethod
    def split_ltm_pool_has_no_avail_mem(msg):
        yield {
            "object": msg.split(" ")[-1],
            "objtype": "pool",
            "new_status": "no members available",
            "prev_status": None,
        }

    @staticmethod
    def split_ltm_pool_has_avail_mem(msg):
        yield {
            "object": msg.split(" ")[1],
            "objtype": "pool",
            "new_status": "now has available members",
            "prev_status": None,
        }

    @staticmethod
    def split_ltm_rule(msg):
        m = F5LogSheet.re_ltm_irule.match(msg)
        if m is None:
            return
        m = m.groupdict()
        yield {
            "object": m.get("irule"),
            "objtype": "rule",
            "irule_event": m.get("event"),
            "irule_msg": m.get("irule_msg"),
            "msg": m.get("message"),
        }
        if m.get("message", "").startswith("aborted for"):
            src = m.get("srchost")
            if src and len(src.split(":")) == 2:
                # ipv4
                srchost, srcport = src.split(":")
            else:
                # ipv6
                srchost, srcport = src.split(".")
            dst = m.get("dsthost")
            if dst and len(dst.split(":")) == 2:
                # ipv4
                dsthost, dstport = dst.split(":")
            else:
                # ipv6
                dsthost, dstport = dst.split(".")
            yield {
                "srchost": ip_address(srchost),
                "srcport": int(srcport),
                "dsthost": ip_address(dsthost),
                "dstport": int(dstport),
            }

    @staticmethod
    def split_ltm_rule_missing_datagroup(msg):
        if "error: Unable to find value_list" in msg:
            m = msg.split(" ", maxsplit=12)
            yield {
                "object": m[1].strip("[").strip("]"),
                "objtype": "rule",
                "msg": "error: Unable to find value_list",
                "missing_dg": m[7].strip("("),
                "funcloc": int(m[11].strip(":")),
                "error": m[12],
            }
        else:
            m = msg.split(" ", maxsplit=5)
            yield {
                "object": m[1].strip("[").strip("]"),
                "objtype": "rule",
                "msg": m[5].split("]", maxsplit=1)[0].strip("[]"),
                "error": m[5].split("]", maxsplit=1)[1],
                "funcloc": int(m[3].split(":")[1]),
            }

    @staticmethod
    def split_ltm_cert_expiry(msg):
        m = F5LogSheet.re_ltm_cert_expiry.match(msg)
        if m is None:
            return
        m = m.groupdict()
        yield {
            "cert_cn": m.get("cert_cn"),
            "object": m.get("file"),
            "objtype": "ssl-cert",
            "date": datetime.strptime(
                m.get("date1").replace("  ", " "),
                "%b %d %H:%M:%S %Y %Z",
            )
            if m.get("date1") is not None
            else None,
            "msg": m.get("message"),
        }

    @staticmethod
    def split_ltm_connection_error(msg):
        m = F5LogSheet.re_ltm_conn_error.match(msg)
        if m is None:
            return
        m = m.groupdict()
        yield {
            "func": m.get("func"),
            "funcloc": m.get("funcloc"),
            "error": m.get("error"),
            "errno": m.get("errno"),
            "errmsg": m.get("errormsg"),
        }

    @staticmethod
    def split_ltm_virtual_status(msg):
        m = msg.split(" ")
        if m[0] == "SNMP_TRAP:":
            yield {
                "object": m[2],
                "objtype": "vs",
                "new_status": m[-1],
                "prev_status": None,
            }
        else:
            yield {
                "object": m[1],
                "objtype": "vs",
                "new_status": m[-1],
                "prev_status": None,
            }

    @staticmethod
    def split_ltm_virtual_address_status_or_irule_profile_err(msg):
        # big-ip has a conflict on this logid
        m = msg.split(" ")
        if "event in rule" in msg:
            yield {
                "object": m[4].strip("()"),
                "objtype": "rule",
                "irule_event": m[0],
                "msg": f"event in rule {' '.join(m[5:-2])}",
                "target_obj": m[-1].strip("()."),
                "target_objtype": m[-2].replace("virtual-server", "vs"),
            }
        else:
            yield {
                "object": m[2],
                "objtype": "virtual address",
                "new_status": m[9].lower().strip("."),
                "prev_status": m[7].lower(),
            }

    @staticmethod
    def split_ltm_ssl_handshake_fail(msg):
        src = msg.split(" ")[5]
        if len(src.split(":")) == 2:
            # ipv4
            srchost, srcport = src.split(":")
        else:
            # ipv6
            srchost, srcport = src.split(".")
        dst = msg.split(" ")[7]
        if len(dst.split(":")) == 2:
            # ipv4
            dsthost, dstport = dst.split(":")
        else:
            dsthost, dstport = dst.split(".")
        yield {
            "srchost": ip_address(srchost),
            "srcport": int(srcport),
            "dsthost": ip_address(dsthost),
            "dstport": int(dstport),
        }

    @staticmethod
    def split_ltm_shared_ciphers(msg):
        m = F5LogSheet.re_ltm_no_shared_ciphers.match(msg)
        if m is None:
            return
        m = m.groupdict()
        yield {
            "srchost": ip_address(m.get("srchost")),
            "srcport": int(m.get("srcport")),
            "dsthost": ip_address(m.get("dsthost")),
            "dstport": int(m.get("dstport")),
        }

    @staticmethod
    def split_ltm_rst_reason(msg):
        m = msg.split(" ", maxsplit=7)
        src, dst = m[3].strip(","), m[5].strip(",")
        reasonc1, reasonc2 = m[6].split(":")
        if len(src.split(":")) == 2:
            # ipv4
            srchost, srcport = src.split(":")
        else:
            # ipv6
            srchost, srcport = src.rsplit(":", maxsplit=1)
        if len(dst.split(":")) == 2:
            # ipv4
            dsthost, dstport = dst.split(":")
        else:
            # ipv6
            dsthost, dstport = dst.rsplit(":", maxsplit=1)
        yield {
            "srchost": ip_address(srchost),
            "srcport": int(srcport) if srcport else None,
            "dsthost": ip_address(dsthost),
            "dstport": int(dstport) if dstport else None,
            "rst_reason_code1": hexint(reasonc1[3:]),
            "rst_reason_code2": hexint(reasonc2[:-1]),
            "rst_reason": m[7],
        }

    @staticmethod
    def split_ltm_inet_port_exhaust(msg):
        m = msg.split(" ")
        srchost, dst = m[-5], m[-3]
        if len(dst.split(":")) == 2:
            dsthost, dstport = dst.split(":")
        else:
            # ipv6
            dsthost, dstport = dst.rsplit(":", maxsplit=1)
        yield {
            "msg": " ".join(m[:5] if len(m) == 11 else m[:3]),
            "srchost": ip_address(srchost),
            "dsthost": ip_address(dsthost),
            "dstport": int(dstport),
            "proto": F5LogSheet._proto[int(m[-1].strip(")"))],
        }

    @staticmethod
    def split_ltm_conn_limit_reached(msg):
        m = msg.split(" ", maxsplit=11)
        src, dst = m[4], m[6].strip(",")
        if len(src.split(":")) == 2:
            srchost, srcport = src.split(":")
        else:
            # ipv6
            srchost, srcport = src.rsplit(".", maxsplit=1)
        if len(dst.split(":")) == 2:
            dsthost, dstport = dst.split(":")
        else:
            # ipv6
            dsthost, dstport = dst.rsplit(".", maxsplit=1)
        yield {
            "msg": m[-1],
            "object": m[10].strip(":"),
            "objtype": m[9].lower(),
            "srchost": ip_address(srchost),
            "srcport": int(dstport),
            "dsthost": ip_address(dsthost),
            "dstport": int(dstport),
            "proto": m[8].strip(",").lower(),
        }

    @staticmethod
    def split_ltm_syncookie_threshold(msg):
        # Syncookie threshold 1993 exceeded, virtual = 203.24.253.132:443
        m = msg.split(" ")
        dst = m[-1]
        if len(dst.split(":")) == 2:
            dsthost, dstport = dst.split(":")
        else:
            # ipv6
            dsthost, dstport = dst.rsplit(".", maxsplit=1)
        yield {
            "msg": "Syncookie threshold exceeded",
            "threshold": int(m[2]),
            "objtype": m[-3].lower(),
            "dsthost": ip_address(dsthost),
            "dstport": int(dstport),
        }

    @staticmethod
    def split_ltm_sweeper_active2(msg):
        m = msg.split(" ")
        yield {
            "policy": m[3],
            "mode": m[4],
            "object": m[7].strip(")."),
            "objtype": m[6].strip("("),
            "msg": " ".join(m[8:]).strip("()"),
        }

    @staticmethod
    def split_ltm_sweeper_active3(msg):
        m = msg.split(" ")
        yield {
            "policy": m[3],
            "object": m[6].strip(")."),
            "objtype": m[5].strip("("),
            "msg": " ".join(m[7:]).strip("()"),
        }

    @staticmethod
    def split_ltm_dns_failed_xfr_rcode(msg):
        m = msg.split(" ")
        yield {
            "msg": " ".join([m[0], *m[4:]]),
            "zone": m[3],
        }

    @staticmethod
    def split_ltm_dns_failed_rr(msg):
        m = msg.rsplit(" ", maxsplit=3)
        yield {
            "msg": m[0],
            "zone": m[-1].strip("."),
        }

    @staticmethod
    def split_ltm_dns_failed_xfr(msg):
        m = msg.split(" ")
        src = m[6].strip(",")
        yield {
            "msg": " ".join(m[:3]) + ", " + " ".join(m[-2:]),
            "srchost": ip_address(src),
            "zone": m[4],
        }

    @staticmethod
    def split_ltm_dns_handling_notify(msg):
        m = msg.split(" ")
        yield {
            "msg": " ".join(m[:2]),
            "zone": m[4].rstrip("."),
        }

    @staticmethod
    def split_ltm_dns_axfr_succeeded_1f(msg):
        m = msg.split(" ")
        yield {
            "msg": " ".join([*m[:3], m[-1]]),
            "srchost": ip_address(m[6]),
            "zone": m[4],
        }

    @staticmethod
    def split_ltm_dns_axfr_succeeded_2c(msg):
        m = msg.split(" ")
        yield {
            "msg": " ".join([*m[:1], m[-1]]),
            "srchost": m[10],
            "zone": m[4],
            "serial": m[8],
        }

    @staticmethod
    def split_ltm_dns_ignoring_tfer(msg):
        # Ignoring transfer for zone qaautomation-dns.com from 10.17.205.164; transfer not enabled.
        m = msg.split(" ")
        yield {
            "msg": " ".join([*m[:2], *m[-3:]]),
            "srchost": m[6].strip(";"),
            "zone": m[4],
        }

    @staticmethod
    def split_ltm_http_process_state(msg):
        m = F5LogSheet.re_ltm_http_process_state.match(msg)
        if not m:
            return
        m = m.groupdict()
        src, vsdst, backendsrc, dst = (
            m.get("src"),
            m.get("vsdst"),
            m.get("poolsrc"),
            m.get("dst"),
        )
        if src:
            if len(src.split(":")) == 2:
                srchost, srcport = src.split(":")
            else:
                # ipv6
                srchost, srcport = src.rsplit(".", maxsplit=1)
        else:
            srchost, srcport = None, None
        if dst:
            if len(dst.split(":")) == 2:
                dsthost, dstport = dst.split(":")
            else:
                # ipv6
                dsthost, dstport = dst.rsplit(".", maxsplit=1)
        else:
            dsthost, dstport = None, None
        if vsdst:
            if len(vsdst.split(":")) == 2:
                vsdsthost, vsdstport = vsdst.split(":")
            else:
                # ipv6
                vsdsthost, vsdstport = vsdst.rsplit(".", maxsplit=1)
        else:
            vsdsthost, vsdstport = None, None
        if backendsrc:
            if len(backendsrc.split(":")) == 2:
                backendsrchost, backendsrcport = backendsrc.split(":")
            else:
                # ipv6
                backendsrchost, backendsrcport = backendsrc.rsplit(".", maxsplit=1)
        else:
            backendsrchost, backendsrcport = None, None
        yield {
            "object": m.get("vs"),
            "msg": m.get("msg"),
            "srchost": ip_address(srchost) if srchost else None,
            "srcport": int(srcport) if srcport else None,
            "dsthost": ip_address(dsthost) if dsthost else None,
            "dstport": int(dstport) if dstport else None,
            "pool": m.get("pool"),
            "profile": m.get("profile"),
            "httpstate": m.get("httpstate"),
            "actionid": hexint(m.get("actionid")),
            "sidea": m.get("sidea"),
            "vsdsthost": ip_address(vsdsthost) if vsdsthost else None,
            "vsdstport": int(vsdstport) if vsdstport else None,
            "backendsrchost": ip_address(backendsrchost) if backendsrchost else None,
            "backendsrcport": int(backendsrcport) if backendsrcport else None,
            "sideb": m.get("sideb"),
            "sideasrc": ip_address(m.get("sideasrc")) if m.get("sideasrc") else None,
        }

    @staticmethod
    def split_ltm_http_header_exceeded(msg):
        m = F5LogSheet.re_ltm_http_header_exceeded.match(msg)
        if not m:
            return
        m = m.groupdict()
        host = m.get("sidehost")
        yield {
            "object": m.get("object"),
            "msg": m.get("msg"),
            "srchost": ip_address(host)
            if host and m.get("sideip") == "client"
            else None,
            "dsthost": ip_address(host)
            if host and m.get("sideip") == "server"
            else None,
            "pool": m.get("pool"),
            "profile": m.get("profile"),
            "size": int(m.get("size")) if m.get("size") else None,
            "limit": int(m.get("limit")) if m.get("limit") else None,
        }

    @staticmethod
    def split_gtm_monitor(msg):
        m = F5LogSheet.re_gtm_monitor.match(msg)
        if m is None:
            return
        m = m.groupdict()
        dst = m.get("ipport")
        if dst:
            if len(dst.split(".")) == 4:
                # ipv4
                if ":" in dst:
                    dsthost, dstport = dst.split(":")
                else:
                    dsthost, dstport = dst, None
            else:
                # ipv6
                if "." in dst:
                    dsthost, dstport = dst.rsplit(".", maxsplit=1)
                else:
                    dsthost, dstport = dst, None
        else:
            dsthost, dstport = None, None
        yield {
            "objtype": m.get("objtype").lower() if m.get("objtype") else None,
            "object": m.get("object"),
            "pool_member": m.get("pool_member"),
            "monitor_object": m.get("monitor_object"),
            "dsthost": ip_address(dsthost) if dsthost else None,
            "dstport": int(dstport) if dstport else None,
            "server": m.get("server"),
            "new_status": m.get("new_status").lower() if m.get("new_status") else None,
            "prev_status": m.get("prev_status").lower()
            if m.get("prev_status")
            else None,
            "msg": m.get("msg"),
            "type": m.get("type").lower() if m.get("type") in m else None,
            "state": m.get("state"),
        }

    @staticmethod
    def split_gtm_monitor_instance(msg):
        m = F5LogSheet.re_gtm_monitor_instance.match(msg)
        if m is None:
            return
        m = m.groupdict()
        if m.get("monip"):
            if len(m.get("monip").split(":")) == 2:
                # ipv4
                dsthost, dstport = m.get("monip").split(":")
            else:
                dsthost, dstport = m.get("monip").rsplit(":", maxsplit=1)
        else:
            dsthost, dstport = None, None
        yield {
            "object": m.get("object"),
            "objtype": "monitor",
            "dsthost": ip_address(dsthost) if dsthost else None,
            "dstport": int(dstport) if dstport else None,
            "new_status": m.get("newstatus", "").lower(),
            "prev_status": m.get("prevstatus", "").lower(),
            "src_gtm": m.get("srcgtm"),
            "state": m.get("state").lower(),
        }

    @staticmethod
    def split_gtm_syncgroup_change(msg):
        m = msg.split(" ")
        yield {
            "object": m[3],
            "srchost": ip_address(m[4].strip("()")),
            "syncgroup": m[-1],
            "msg": f"BIG-IP GTM {m[5]} sync group",
        }

    @staticmethod
    def split_gtm_changed_state(msg):
        m = msg.split(" ")
        yield {
            "msg": f"{m[0]} changed state",
            "new_status": m[6].lower().strip("."),
            "prev_status": m[4].lower(),
        }

    @staticmethod
    def split_tmm_address_conflict(msg):
        m = msg.split(" ")
        dsthost = m[4]
        yield {
            "object": " ".join(m[7:]),
            "objtype": "address",
            "dsthost": ip_address(dsthost),
            "dstmac": m[5].strip("()"),
        }

    splitters = {
        0x01010028: split_ltm_pool_has_no_avail_mem.__func__,
        0x01010038: split_ltm_syncookie_threshold.__func__,
        0x01010201: split_ltm_inet_port_exhaust.__func__,
        0x01010281: split_ltm_inet_port_exhaust.__func__,
        0x01010221: split_ltm_pool_has_avail_mem.__func__,
        0x01070151: split_ltm_rule_missing_datagroup.__func__,
        0x01070417: split_audit_mcpd_mcp_error.__func__,
        0x01070639: split_ltm_poolnode_mon_abled.__func__,
        0x01070641: split_ltm_poolnode_mon_abled.__func__,
        0x01070807: split_ltm_poolnode_mon_abled.__func__,
        0x01070808: split_ltm_poolnode_mon_abled.__func__,
        0x01070727: split_ltm_pool_mon_status.__func__,
        0x01070728: split_ltm_pool_mon_status.__func__,
        0x01070638: split_ltm_pool_mon_status.__func__,
        0x01070640: split_ltm_pool_mon_status.__func__,
        0x01071681: split_ltm_virtual_status.__func__,
        0x01071682: split_ltm_virtual_status.__func__,
        0x01071912: split_ltm_virtual_address_status_or_irule_profile_err.__func__,
        0x01071913: split_ltm_virtual_address_status_or_irule_profile_err.__func__,
        0x010719E7: split_ltm_virtual_address_status_or_irule_profile_err.__func__,
        0x010719E8: split_ltm_virtual_address_status_or_irule_profile_err.__func__,
        0x010719EA: split_gtm_changed_state.__func__,
        0x01071BA9: split_ltm_virtual_status.__func__,
        0x01190004: split_tmm_address_conflict.__func__,
        0x011A1004: split_gtm_monitor.__func__,
        0x011A1005: split_gtm_monitor.__func__,
        0x011A1101: split_gtm_monitor.__func__,
        0x011A1102: split_gtm_monitor.__func__,
        0x011A3003: split_gtm_monitor.__func__,
        0x011A3004: split_gtm_monitor.__func__,
        0x011A4002: split_gtm_monitor.__func__,
        0x011A4003: split_gtm_monitor.__func__,
        0x011A4004: split_gtm_monitor.__func__,
        0x011A4101: split_gtm_monitor.__func__,
        0x011A4102: split_gtm_monitor.__func__,
        0x011A5003: split_gtm_monitor.__func__,
        0x011A5004: split_gtm_monitor.__func__,
        0x011A5008: split_gtm_syncgroup_change.__func__,
        0x011A5009: split_gtm_syncgroup_change.__func__,
        0x011A500B: split_gtm_monitor.__func__,
        0x011A500C: split_gtm_monitor.__func__,
        0x011A6005: split_gtm_monitor.__func__,
        0x011A6006: split_gtm_monitor.__func__,
        0x011AB003: split_gtm_monitor.__func__,
        0x011AB004: split_gtm_monitor.__func__,
        0x011AE0F2: split_gtm_monitor_instance.__func__,
        # 0x01220000: split_ltm_rule.__func__,
        0x011E0002: split_ltm_sweeper_active2.__func__,
        0x011E0003: split_ltm_sweeper_active3.__func__,
        0x011F0005: split_ltm_http_header_exceeded.__func__,
        0x011F0011: split_ltm_http_header_exceeded.__func__,
        0x011F0007: split_ltm_http_process_state.__func__,
        0x011F0016: split_ltm_http_process_state.__func__,
        0x01220001: split_ltm_rule.__func__,
        0x01220002: split_ltm_rule.__func__,
        # 0x01220003: split_ltm_rule.__func__,
        # 0x01220004: split_ltm_rule.__func__,
        # 0x01220005: split_ltm_rule.__func__,
        0x01220007: split_ltm_rule.__func__,
        0x01220008: split_ltm_rule.__func__,
        0x01220009: split_ltm_rule.__func__,
        0x01220010: split_ltm_rule.__func__,
        0x01220011: split_ltm_rule.__func__,
        0x01200012: split_ltm_conn_limit_reached.__func__,
        0x01200014: split_ltm_conn_limit_reached.__func__,
        0x01230140: split_ltm_rst_reason.__func__,
        0x01260013: split_ltm_ssl_handshake_fail.__func__,
        0x01260026: split_ltm_shared_ciphers.__func__,
        0x01260008: split_ltm_connection_error.__func__,
        0x01260009: split_ltm_connection_error.__func__,
        0x01420002: split_audit_bigip_tmsh_audit.__func__,
        0x01420007: split_ltm_cert_expiry.__func__,
        0x01420008: split_ltm_cert_expiry.__func__,
        0x014F0005: split_audit_scriptd_run_script.__func__,
        0x0153100E: split_ltm_dns_failed_xfr_rcode.__func__,
        0x01531015: split_ltm_dns_failed_rr.__func__,
        0x01531018: split_ltm_dns_failed_xfr.__func__,
        0x0153101C: split_ltm_dns_handling_notify.__func__,
        0x0153101F: split_ltm_dns_axfr_succeeded_1f.__func__,
        0x01531022: split_ltm_dns_ignoring_tfer.__func__,
        0x0153102C: split_ltm_dns_axfr_succeeded_2c.__func__,
    }

    # these logs can have IDs we care about splitting but would be errors
    # the match is starts_with because of course some logs have extra dynamic junk
    no_split_logs = (
        "Per-invocation log rate exceeded; throttling",
        "Resuming log processing at this invocation",
        "Re-enabling general logging;",
        "Cumulative log rate exceeded!  Throttling all non-debug logs.",
    )

    extra_cols = {
        "rawmsg",
        "timestamp",
        "host",
        "level",
        "process",
        "proc_pid",
        "logid1",
        "logid2",
        "message",
        "object",
    }

    # precedence, coloropt, func
    colorizers = [
        CellColorizer(100, None, colorizeMonitors),
        RowColorizer(101, None, colorizeRows),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # the default F5 logs don't have the year so we have to guess from the file ctime
        # TODO: make this overridable
        try:
            import zoneinfo
        except ImportError:
            from backports import zoneinfo
        self._log_tz = zoneinfo.ZoneInfo("UTC")
        try:
            self._year = int(
                vd.options.get(
                    "f5log_log_year",
                    datetime.utcfromtimestamp(self.source.stat().st_ctime).year,
                )
            )
        except (AttributeError, ValueError, TypeError):
            self._year = datetime.now().year

    def iterload(self):
        self.rows = []  # rowdef: [F5LogRow]

        if vd.options.get("f5log_object_regex"):
            try:
                object_regex = re.compile(vd.options.get("f5log_object_regex"))
            except re.error as exc:
                # TODO: make this error into the errors sheet
                object_regex = None
        else:
            object_regex = None

        try:
            import zoneinfo
            self._log_tz = zoneinfo.ZoneInfo(
                vd.options.get("f5log_log_timzeone", "UTC")
            )
        except zoneinfo.ZoneInfoNotFoundError as exc:
            # TODO: make this error go into the errors sheet
            self._log_tz = zoneinfo.ZoneInfo("UTC")

        for line in self.source:
            m = F5LogSheet.re_f5log.match(line)
            if m:
                m = m.groupdict()
            else:
                # TODO: somehow make this use an error sheet
                yield F5LogSheet.F5LogRow(
                    rawmsg=line, kv={"PARSE_ERROR": "unable to parse line"}
                )
                continue
            kv = {
                "message": m.get("message"),
            }
            if m.get("date1"):
                #
                _t = m.get("date1")
                # strptime is quite slow so we need to manually extract the time on the hot path
                try:
                    timestamp = datetime(
                        year=self._year,
                        month=self._months[_t[:3]],
                        day=int(_t[4:6]),
                        hour=int(_t[7:9]),
                        minute=int(_t[10:12]),
                        second=int(_t[13:15]),
                        tzinfo=self._log_tz,
                    )
                except ValueError as exc:
                    yield F5LogSheet.F5LogRow(
                        rawmsg=line,
                        PARSE_ERROR="\n".join(
                            traceback.format_exception(
                                etype=type(exc), value=exc, tb=exc.__traceback__
                            ),
                        ),
                    )
            elif m.get("date2"):
                timestamp = datetime.strptime(m.get("date2"), "%Y-%m-%dT%H:%M:%S%z")
            elif m.get("date3"):
                # whoever designed tmsh show sys log needs to have a good hard think about themselves
                timestamp = datetime.strptime(
                    f'{self._year}-{m.get("date3")}', "%Y-%m-%d %H:%M:%S"
                )
                timestamp = datetime(
                    year=timestamp.year,
                    month=timestamp.month,
                    day=timestamp.day,
                    hour=timestamp.hour,
                    minute=timestamp.minute,
                    second=timestamp.second,
                    tzinfo=self._log_tz,
                )
                # because this is madness
                m["level"], m["host"] = m.get("host"), m.get("level")
            else:
                timestamp = None

            logid1 = int(m.get("logid1"), base=16) if m.get("logid1") else None
            if logid1 in self.splitters and not any(
                m.get("message", "").startswith(_) for _ in F5LogSheet.no_split_logs
            ):
                try:
                    for entry in F5LogSheet.splitters[logid1](m.get("message")):
                        kv.update(entry)
                except (IndexError, ValueError) as exc:
                    # TODO: somehow make this use an error sheet
                    yield F5LogSheet.F5LogRow(
                        rawmsg=line,
                        PARSE_ERROR="\n".join(
                            traceback.format_exception(
                                etype=type(exc), value=exc, tb=exc.__traceback__
                            )
                        ),
                    )
                if "object" in kv and object_regex:
                    om = object_regex.match(kv.get("object", ""))
                    if om:
                        kv.update(om.groupdict())
                for k, v in kv.items():
                    if k not in self.extra_cols:
                        F5LogSheet.addColumn(self, AttrColumn(k))
                        self.extra_cols.add(k)
            elif logid1 is None and m.get("message").startswith("Rule "):
                for entry in self.split_ltm_rule(m.get("message")):
                    kv.update(entry)
            yield F5LogSheet.F5LogRow(
                # rawmsg=line,
                timestamp=timestamp,
                host=m.get("host"),
                level=m.get("level"),
                process=m.get("process"),
                proc_pid=int(m.get("pid")) if m.get("pid") is not None else None,
                logid1=m.get("logid1") if m.get("logid1") is not None else None,
                logid2=m.get("logid2") if m.get("logid2") is not None else None,
                **kv,
            )


@VisiData.api
def open_f5log(vd: VisiData, p: Path) -> Sheet:
    sheet = F5LogSheet(p.base_stem, source=p)
    sheet.options["disp_date_fmt"] = "%Y-%m-%d %H:%M:%S"
    return sheet



from collections import defaultdict

from visidata import Sheet, VisiData, TypedWrapper, anytype, date, vlen, Column, vd



@VisiData.api
def open_arrow(vd, p):
    'Apache Arrow IPC file format'
    return ArrowSheet(p.base_stem, source=p)


@VisiData.api
def open_arrows(vd, p):
    'Apache Arrow IPC streaming format'
    return ArrowSheet(p.base_stem, source=p)


def arrow_to_vdtype(t):
    pa = vd.importExternal('pyarrow')

    arrow_to_vd_typemap = {
        pa.lib.Type_BOOL: bool,
        pa.lib.Type_UINT8: int,
        pa.lib.Type_UINT16: int,
        pa.lib.Type_UINT32: int,
        pa.lib.Type_UINT64: int,
        pa.lib.Type_INT8: int,
        pa.lib.Type_INT16: int,
        pa.lib.Type_INT32: int,
        pa.lib.Type_INT64: int,
        pa.lib.Type_HALF_FLOAT: float,
        pa.lib.Type_FLOAT: float,
        pa.lib.Type_DOUBLE: float,
#            pa.lib.Type_DECIMAL128: Decimal128Scalar,
#            pa.lib.Type_DECIMAL256: Decimal256Scalar,
        pa.lib.Type_DATE32: date,
        pa.lib.Type_DATE64: date,
        pa.lib.Type_TIME32: date,
        pa.lib.Type_TIME64: date,
        pa.lib.Type_TIMESTAMP: date,
        pa.lib.Type_DURATION: int,
        pa.lib.Type_BINARY: bytes,
        pa.lib.Type_LARGE_BINARY: vlen,
#            pa.lib.Type_FIXED_SIZE_BINARY: bytes,
#            pa.lib.Type_STRING: str,
#            pa.lib.Type_LARGE_STRING: vlen,  #2003
#            pa.lib.Type_LIST: list,
#            pa.lib.Type_LARGE_LIST: list,
#            pa.lib.Type_FIXED_SIZE_LIST: list,
#            pa.lib.Type_STRUCT: dict,
#            pa.lib.Type_MAP: dict,
#            pa.lib.Type_DICTIONARY: dict,
#            pa.lib.Type_SPARSE_UNION: UnionScalar,
#            pa.lib.Type_DENSE_UNION: UnionScalar,
    }
    return arrow_to_vd_typemap.get(t.id, anytype)

class ArrowSheet(Sheet):
    def iterload(self):
        pa = vd.importExternal('pyarrow')

        try:
            with pa.OSFile(str(self.source), 'rb') as fp:
                self.coldata = pa.ipc.open_file(fp).read_all()
        except pa.lib.ArrowInvalid as e:
            with pa.OSFile(str(self.source), 'rb') as fp:
                self.coldata = pa.ipc.open_stream(fp).read_all()

        self.columns = []
        for colnum, col in enumerate(self.coldata):
            coltype = arrow_to_vdtype(self.coldata.schema.types[colnum])
            colname = self.coldata.schema.names[colnum]

            self.addColumn(Column(colname, type=coltype, expr=colnum,
                                  getter=lambda c,r: c.sheet.coldata[c.expr][r[0]].as_py()))

        for rownum in range(max(len(c) for c in self.coldata)):
            yield [rownum]


@VisiData.api
def save_arrow(vd, p, sheet, streaming=False):
    pa = vd.importExternal('pyarrow')
    np = vd.importExternal('numpy')

    typemap = {
        anytype: pa.string(),
        int: pa.int64(),
        vlen: pa.int64(),
        float: pa.float64(),
        str: pa.string(),
        date: pa.date64(),
    }

    for t in vd.numericTypes:
        if t not in typemap:
            typemap[t] = pa.float64()

    databycol = defaultdict(list)   # col -> [values]

    for typedvals in sheet.iterdispvals(format=False):
        for col, val in typedvals.items():
            if isinstance(val, TypedWrapper):
                val = None

            databycol[col].append(val)

    data = [pa.array(vals, type=typemap.get(col.type, pa.string())) for col, vals in databycol.items()]

    schema = pa.schema([
        (c.name, typemap.get(c.type, pa.string()))
            for c in sheet.visibleCols
    ])
    with p.open_bytes(mode='w') as outf:
        if streaming:
            with pa.ipc.new_stream(outf, schema) as writer:
                writer.write_batch(pa.record_batch(data, names=[c.name for c in sheet.visibleCols]))
        else:
            with pa.ipc.new_file(outf, schema) as writer:
                writer.write_batch(pa.record_batch(data, names=[c.name for c in sheet.visibleCols]))


@VisiData.api
def save_arrows(vd, p, sheet):
    return vd.save_arrow(p, sheet, streaming=True)



import collections

from visidata import vd, BaseSheet


@BaseSheet.lazy_property
def prevHints(sheet):
    return collections.defaultdict(int)


@BaseSheet.api
def getHint(sheet, *args, **kwargs) -> str:
    funcs = [getattr(sheet, x) for x in dir(sheet) if x.startswith('hint_')]
    results = []
    hints = sheet.prevHints
    for f in funcs:
        try:
            r = f(*args, **kwargs)
            if r:
                if isinstance(r, dict):
                    n = r.get('_relevance', 1)
                    v = r
                elif isinstance(r, tuple):
                    n, v = r
                else:
                    n = 1
                    v = r

                results.append((n, v))
                if v not in sheet.prevHints:
                    sheet.prevHints[v] += 1
        except Exception as e:
            vd.debug(f'{f.__name__}: {e}')

    if results:
        return sorted(results, reverse=True)[0][1]

    return ''

vd.addCommand('', 'help-hint', 'status(getHint() or pressMenu("Help"))', 'get context-dependent hint on what to do next')



import re

from visidata import vd, Column, VisiData, ItemColumn, AttrColumn, Path, AttrDict, RowColorizer, date, Progress

from .gitsheet import GitSheet

vd.theme_option('color_git_current_branch', 'underline', 'color of current branch on branches sheet')
vd.theme_option('color_git_remote_branch', 'cyan', 'color of remote branches on branches sheet')


@VisiData.api
def git_branch(vd, p, args):
    nonListArgs = '--track --no-track --set-upstream-to -u --unset-upstream -m -M -c -C -d -D --edit-description'.split()
    if any(x in args for x in nonListArgs):
        return

    return GitBranch('git-branch-list', source=p, git_args=args)


def _remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


class GitBranchNameColumn(Column):
    def calcValue(self, row):
        return _remove_prefix(row.localbranch, 'remotes/')

    def putValue(self, row, val):
        self.sheet.loggit('branch', '-v', '--move', row.localbranch, val)


class GitBranch(GitSheet):
    guide = '''
        # git branch
        List of all branches, including relevant metadata.

        - `d` to mark a branch for deletion
        - `e` on the _branch_ column to rename the branch
        - `z Ctrl+S` to commit changes
    '''
    defer = True
    rowtype = 'branches'  # rowdef: AttrDict from regex (in reload below)
    columns = [
        GitBranchNameColumn('branch', width=20),
#        Column('remote', getter=lambda c,r: r['localbranch'].startswith('remotes/') and '*' or '', width=3),
        ItemColumn('head_commitid', 'refid', width=0),
        ItemColumn('tracking', 'remotebranch'),
        ItemColumn('upstream'),
        ItemColumn('merge_base', 'merge_name', width=20),
        ItemColumn('extra', width=0),
        ItemColumn('head_commitmsg', 'msg', width=50),
        ItemColumn('last_commit', type=date),
        ItemColumn('last_author'),
    ]
    colorizers = [
        RowColorizer(10, 'color_git_current_branch', lambda s,c,r,v: r and r['current']),
        RowColorizer(10, 'color_git_remote_branch', lambda s,c,r,v: r and r['localbranch'].startswith('remotes/')),
    ]
    nKeys = 1

    def iterload(self):
        branches_lines = self.git_lines(
            'branch',
            '--list',
            '--format',
            ' '.join((
                '%(if)%(symref)%(then)yes%(else)no%(end)',
                '%(HEAD) %(refname:short) %(objectname:short)',
                '%(if)%(upstream)%(then)[%(upstream:short)',
                '%(if)%(upstream:track)%(then): %(upstream:track,nobracket)%(end)]',
                '%(end)',
                '%(contents:subject)'
            )),
            '-vv',
            '--no-color',
            *self.git_args)

        for line in branches_lines:
            m = re.match(r'''(?P<is_symref>(yes|no)?)\s+
                             (?P<current>\*?)\s+
                             (?P<localbranch>\S+)\s+
                             (?P<refid>\w+)\s+
                             (?:\[
                               (?P<remotebranch>[^\s\]:]+):?
                               \s*(?P<extra>.*?)
                             \])?
                             \s*(?P<msg>.*)''', line, re.VERBOSE)
            if not m:
                continue
            branch_details = AttrDict(m.groupdict())
            if branch_details.is_symref == 'yes':
                continue
            yield branch_details

        branch_stats = self.gitRootSheet.gitBranchStatuses
        for row in Progress(self.rows):
            merge_base = self.git_all("show-branch", "--merge-base", row.localbranch, self.gitRootSheet.branch, _ok_code=[0,1]).strip()
            row.update(dict(
                merge_name = self.git_all("name-rev", "--name-only", merge_base).strip() if merge_base else '',
                upstream = branch_stats.get(row.localbranch),
                last_commit = self.git_all("show", "--no-patch", '--pretty=%ai', row.localbranch).strip(),
                last_author = self.git_all("show", "--no-patch", '--pretty=%an', row.localbranch).strip()
            ))

    def commitAddRow(self, row):
        self.loggit('branch', row.localbranch)

    def commitDeleteRow(self, row):
        self.loggit('branch', '--delete', _remove_prefix(row.localbranch, 'remotes/'))


@GitSheet.lazy_property
def gitBranchStatuses(sheet):
    ret = {}  # localbranchname -> "+5/-2"
    for branch_status in sheet.git_lines('for-each-ref', '--format=%(refname:short) %(upstream:short) %(upstream:track)', 'refs/heads'):
        m = re.search(r'''(\S+)\s*
                          (\S+)?\s*
                          (\[
                          (ahead.(\d+)),?\s*
                          (behind.(\d+))?
                          \])?''', branch_status, re.VERBOSE)
        if not m:
            vd.status('unmatched branch status: ' + branch_status)
            continue

        localb, remoteb, _, _, nahead, _, nbehind = m.groups()
        if nahead:
            r = '+%s' % nahead
        else:
            r = ''
        if nbehind:
            if r:
                r += '/'
            r += '-%s' % nbehind
        ret[localb] = r

    return ret


GitSheet.addCommand('', 'git-open-branches', 'vd.push(git_branch(source, []))', 'push branches sheet')
GitSheet.addCommand('', 'git-branch-create', 'git("branch", input("create branch: ", type="branch"))', 'create a new branch off the current checkout')
GitBranch.addCommand('', 'git-branch-checkout', 'git("checkout", cursorRow.localbranch)', 'checkout this branch')


vd.addMenuItems('''
    Git > Branch > add > git-branch-create
    Git > Branch > delete > git-branch-delete
    Git > Branch > rename > git-branch-rename
    Git > Branch > checkout > git-branch-checkout
    Git > Open > branches > git-open-branches
''')



import re
import random

from visidata import asyncthread, options, vd
from visidata import VisiData, BaseSheet, Sheet, Column, Progress


@VisiData.lazy_property
def help_regex(vd):
    return vd.getHelpPane('regex', module='visidata')


vd.option('regex_maxsplit', 0, 'maxsplit to pass to regex.split', replay=True)

@VisiData.api
def makeRegexSplitter(vd, regex, origcol):
    return lambda row, regex=regex, origcol=origcol, maxsplit=options.regex_maxsplit: regex.split(origcol.getDisplayValue(row), maxsplit=maxsplit)

@VisiData.api
def makeRegexMatcher(vd, regex, origcol):
    if not regex.groups:
        vd.fail('specify a capture group')  #1778
    def _regexMatcher(row):
        m = regex.search(origcol.getDisplayValue(row))
        if m:
            return m.groupdict() if m.groupdict() else m.groups()
    return _regexMatcher


@Sheet.api
def RegexColumn(vs, regexMaker, origcol, regexstr):
    regex = re.compile(regexstr, vs.regex_flags())
    func = regexMaker(regex, origcol)
    return Column(origcol.name+'_re',
                  getter=lambda col,row,func=func: func(row),
                  origCol=origcol)


@Sheet.api
@asyncthread
def addRegexColumns(vs, regexMaker, origcol, regexstr):
    regexstr or vd.fail('regex required')

    regex = re.compile(regexstr, vs.regex_flags())

    func = regexMaker(regex, origcol)

    cols = {}
    ncols = 0  # number of new columns added already
    for r in Progress(vs.getSampleRows()):
        m = vd.callNoExceptions(func, r)
        if not m:
            continue

        if isinstance(m, dict):
            for name in m:
                if name in cols:
                    continue
                cols[name] = Column(origcol.name+'_'+str(name),
                                    getter=lambda col,row,name=name,func=func: func(row)[name],
                                    origCol=origcol)
        elif isinstance(m, (tuple, list)):
            for _ in range(len(m)-len(cols)):
                cols[len(cols)] = Column(origcol.name+'_re'+str(len(cols)),
                                         getter=lambda col,row,i=len(cols),func=func: func(row)[i],
                                         origCol=origcol)
        else:
            raise TypeError("addRegexColumns() expects a dict, list, or tuple from regexMaker, but got a "+type(m).__name__)

    if not cols:
        vd.warning("no regex matches found, didn't add column")
        return

    vs.addColumnAtCursor(*cols.values())


@VisiData.api
def regexTransform(vd, origcol, before='', after=''):
    return lambda col,row,origcol=origcol,before=before,after=after,flags=origcol.sheet.regex_flags(): re.sub(before, after, origcol.getDisplayValue(row), flags=flags)


@VisiData.api
def parse_sed_transform(vd, instr):
    i = indexWithEscape(instr, '/')
    if i is None:
        return instr, ''
    else:
        return instr[:i], instr[i+1:]


def indexWithEscape(s, char, escape_char='\\'):
    i=0
    while i < len(s):
        if s[i] == escape_char:
            i += 1
        elif s[i] == char:
            return i
        i += 1

    return None


@Sheet.api
@asyncthread
def setValuesFromRegex(sheet, cols, rows, before='', after=''):
    transforms = [vd.regexTransform(col, before=before, after=after) for col in cols]
    vd.addUndoSetValues(cols, rows)
    for r in Progress(rows, 'replacing'):
        for col, transform in zip(cols, transforms):
            vd.callNoExceptions(col.setValue, r, transform(col, r))

    for col in cols:
        col.recalc()


@VisiData.api
def inputRegex(vd, prompt, type='regex', **kwargs):
    return vd.input(prompt, type=type, help=vd.help_regex, **kwargs)



@VisiData.api
def inputRegexSubst(vd, prompt):
    'Input regex transform via oneliner (separated with `/`).  Return parsed transformer as dict(before=, after=).'
    return vd.inputMultiple(before=dict(type='regex', prompt='search: ', help=prompt),
                            after=dict(type='regex-replace', prompt='replace: ', help=prompt))


Sheet.addCommand(':', 'addcol-split', 'addColumnAtCursor(RegexColumn(makeRegexSplitter, cursorCol, inputRegex("split regex: ", type="regex-split")))', 'add column split by regex')
Sheet.addCommand(';', 'addcol-capture', 'addColumnAtCursor(RegexColumn(makeRegexMatcher, cursorCol, inputRegex("capture regex: ", type="regex-capture")))', 'add column captured by regex')

Sheet.addCommand('*', 'addcol-regex-subst', 'addColumnAtCursor(Column(cursorCol.name + "_re", getter=regexTransform(cursorCol, **inputRegexSubst("regex transform column"))))', 'add column derived from current column, replacing `search` regex with `replace` (may include \\1 backrefs)')
Sheet.addCommand('g*', 'setcol-regex-subst', 'setValuesFromRegex([cursorCol], someSelectedRows, **inputRegexSubst("regex transform column"))', 'modify selected rows in current column, replacing `search` regex with `replace`, (may include backreferences \\1 etc)')
Sheet.addCommand('gz*', 'setcol-regex-subst-all', 'setValuesFromRegex(visibleCols, someSelectedRows, **inputRegexSubst(f"regex transform {nVisibleCols} columns"))', 'modify selected rows in all visible columns, replacing `search` regex with `replace` (may include \\1 backrefs)')


vd.addMenuItems('''
    Edit > Modify > selected cells > regex substitution > setcol-regex-subst
    Column > Add column > capture by regex > addcol-capture
    Column > Add column > split by regex > addcol-split
    Column > Add column > subst by regex > addcol-regex-subst
    Row > Select > by regex > current column > select-col-regex
    Row > Select > by regex > all columns > select-cols-regex
    Row > Unselect > by regex > current column > unselect-col-regex
    Row > Unselect > by regex > all columns > unselect-cols-regex
''')



from visidata import vd, VisiData, SequenceSheet, options, stacktrace
from visidata import TypedExceptionWrapper, Progress

vd.option('csv_dialect', 'excel', 'dialect passed to csv.reader', replay=True)
vd.option('csv_delimiter', ',', 'delimiter passed to csv.reader', replay=True)
vd.option('csv_doublequote', True, 'quote-doubling setting passed to csv.reader', replay=True)
vd.option('csv_quotechar', '"', 'quotechar passed to csv.reader', replay=True)
vd.option('csv_quoting', 0, 'quoting style passed to csv.reader and csv.writer', replay=True)
vd.option('csv_skipinitialspace', True, 'skipinitialspace passed to csv.reader', replay=True)
vd.option('csv_escapechar', None, 'escapechar passed to csv.reader', replay=True)
vd.option('csv_lineterminator', '\r\n', 'lineterminator passed to csv.writer', replay=True)
vd.option('safety_first', False, 'sanitize input/output to handle edge cases, with a performance cost', replay=True)


@VisiData.api
def guess_csv_delimiter(vd, p):
    'If csv_delimiter option has been modified from default, assume CSV format.'
    
    if vd.options.csv_delimiter != vd.options.getdefault('csv_delimiter'):
        return dict(filetype='csv', _likelihood=2)

@VisiData.api
def guess_csv(vd, p):
    import csv
    csv.field_size_limit(2**31-1)  #288 Windows has max 32-bit
    try:
        line = next(p.open())
    except StopIteration:
        return
    if ',' in line:
        dialect = csv.Sniffer().sniff(line)
        r = dict(filetype='csv', _likelihood=0)

        for csvopt in dir(dialect):
            if not csvopt.startswith('_'):
                v = getattr(dialect, csvopt)
                optname = 'csv_'+csvopt
                r[optname] = v
                if vd.options.get(optname) != v:
                    vd.warning(f'guessed option {optname}={v}')

        return r

@VisiData.api
def open_csv(vd, p):
    return CsvSheet(p.base_stem, source=p)

def removeNulls(fp):
    for line in fp:
        yield line.replace('\0', '')

class CsvSheet(SequenceSheet):
    _rowtype = list  # rowdef: list of values

    def iterload(self):
        'Convert from CSV, first handling header row specially.'
        import csv
        csv.field_size_limit(2**31-1)  #288 Windows has max 32-bit

        with self.open_text_source(newline='') as fp:
            if options.safety_first:
                rdr = csv.reader(removeNulls(fp), **options.getall('csv_'))
            else:
                rdr = csv.reader(fp, **options.getall('csv_'))

            while True:
                try:
                    yield next(rdr)
                except csv.Error as e:
                    e.stacktrace=stacktrace()
                    yield [TypedExceptionWrapper(None, exception=e)]
                except StopIteration:
                    return


@VisiData.api
def save_csv(vd, p, sheet):
    'Save as single CSV file, handling column names as first line.'
    import csv
    csv.field_size_limit(2**31-1)  #288 Windows has max 32-bit

    with p.open(mode='w', encoding=sheet.options.save_encoding, newline='') as fp:
        cw = csv.writer(fp, **options.getall('csv_'))
        colnames = [col.name for col in sheet.visibleCols]
        if ''.join(colnames):
            cw.writerow(colnames)

        with Progress(gerund='saving', total=sheet.nRows) as prog:
            for dispvals in sheet.iterdispvals(format=True):
                cw.writerow(dispvals.values())
                prog.addProgress(1)

vd.addGlobals({
    'CsvSheet': CsvSheet
})



import sys
import math
import functools
import collections
import statistics
from copy import copy
import itertools

from visidata import Progress, Sheet, Column, ColumnsSheet, VisiData, SettableColumn
from visidata import vd, anytype, vlen, asyncthread, wrapply, AttrDict, date, INPROGRESS, dispwidth, stacktrace, TypedExceptionWrapper

vd.help_aggregators = '''# Choose Aggregators
Start typing an aggregator name or description.
Multiple aggregators can be added by separating spaces.

- `Enter` to select top aggregator.
- `Tab` to highlight top aggregator.

## When Aggregator Highlighted

- `Tab`/`Shift+Tab` to cycle highlighted aggregator.
- `Enter` to select aggregators.
- `Space` to add highlighted aggregator.
- `0-9` to add numbered aggregator.
'''

vd.option('null_value', None, 'a value to be counted as null', replay=True)


@Column.api
def getValueRows(self, rows):
    'Generate (value, row) for each row in *rows* at this column, excluding null and error values.'
    f = self.sheet.isNullFunc()

    for r in Progress(rows, 'calculating'):
        try:
            v = self.getTypedValue(r)
            if not f(v):
                yield v, r
        except Exception:
            pass

@Column.api
def getValues(self, rows):
    'Generate value for each row in *rows* at this column, excluding null and error values.'
    for v, r in self.getValueRows(rows):
        yield v


vd.aggregators = collections.OrderedDict()  # [aggname] -> annotated func, or list of same

Column.init('aggstr', str, copy=True)
Column.init('_aggregatedTotals', dict)  # [aggname] -> agg total over all rows

def aggregators_get(col):
    'A space-separated names of aggregators on this column.'
    aggs = []
    for k in (col.aggstr or '').split():
        agg = vd.aggregators[k]
        aggs += agg if isinstance(agg, list) else [agg]
    return aggs

def aggregators_set(col, aggs):
    if isinstance(aggs, str):
        newaggs = []
        for agg in aggs.split():
            if agg not in vd.aggregators:
                vd.fail(f'unknown aggregator {agg}')
            newaggs.append(agg)
    elif aggs is None:
        newaggs = ''
    else:
        newaggs = [agg.name for agg in aggs]

    col.aggstr = ' '.join(newaggs)

Column.aggregators = property(aggregators_get, aggregators_set)


class Aggregator:
    def __init__(self, name, type, funcValues=None, helpstr=''):
        'Define aggregator `name` that calls funcValues(values)'
        self.type = type
        self.funcValues = funcValues  # funcValues(values)
        self.helpstr = helpstr
        self.name = name

    def aggregate(self, col, rows):  # wrap builtins so they can have a .type
        vals = list(col.getValues(rows))
        try:
            return self.funcValues(vals)
        except Exception as e:
            if len(vals) == 0:
                return None
            raise e

class ListAggregator(Aggregator):
    '''A list aggregator is an aggregator that returns a list of values, generally
    one value per input row, unlike ordinary aggregators that operate on rows
    and return only a single value.
    To implement a new list aggregator, subclass ListAggregator,
    and override aggregate() and aggregate_list().'''
    def __init__(self, name, type, helpstr='', listtype=None):
        '''*listtype* determines the type of the column created by addcol_aggregate()
        for list aggrs. If it is None, then the new column will match the type of the input column'''
        super().__init__(name, type, helpstr=helpstr)
        self.listtype = listtype

    def aggregate(self, col, rows) -> list:
        '''Return a list, which can be shorter than *rows*, because it filters out nulls and errors.
        Override in subclass.'''
        vals = self.aggregate_list(col, rows)
        # filter out nulls and errors
        vals = [ v for v in vals if not col.sheet.isNullFunc()(v) ]
        return vals

    def aggregate_list(self, col, row_group) -> list:
        '''Return a list of results, which will be one result per input row.
        *row_group* is an iterable that holds a "group" of rows to run the aggregator on.
        rows in *row_group* are not necessarily in the same order they are in the sheet.
        Override in subclass.'''
        vals = [ col.getTypedValue(r) for r in row_group ]
        return vals

@VisiData.api
def aggregator(vd, name, funcValues, helpstr='', *, type=None):
    '''Define simple aggregator *name* that calls ``funcValues(values)`` to aggregate *values*.
       Use *type* to force type of aggregated column (default to use type of source column).'''
    vd.aggregators[name] = Aggregator(name, type, funcValues=funcValues, helpstr=helpstr)

@VisiData.api
def aggregator_list(vd, name, helpstr='', type=anytype, listtype=anytype):
    '''Define simple aggregator *name* that calls ``funcValues(values)`` to aggregate *values*.
       Use *type* to force type of aggregated column (default to use type of source column).
       Use *listtype* to force the type of the new column created by addcol-aggregate.
       If *listtype* is None, it will match the type of the source column.'''
    vd.aggregators[name] = ListAggregator(name, type, helpstr=helpstr, listtype=listtype)

## specific aggregator implementations

def mean(vals):
    vals = list(vals)
    if vals:
        return float(sum(vals))/len(vals)

def vsum(vals):
    return sum(vals, start=type(vals[0] if len(vals) else 0)())  #1996

def stdev(vals):
    # because statistics.stdev can raise an exception, we put it in a wrapper.
    # The wrapper lets the exception be seen as an error string in the stdev
    # aggregator, shown at the bottom of the sheet as part of allAggregators.
    try:
        return statistics.stdev(vals)
    except statistics.StatisticsError as e:  #when vals holds only 1 element
        e.stacktrace = stacktrace()
        return TypedExceptionWrapper(None, exception=e)

# http://code.activestate.com/recipes/511478-finding-the-percentile-of-the-values/
def _percentile(N, percent, key=lambda x:x):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values. Note N MUST BE already sorted.
    @parameter percent - a float value from 0.0 to 1.0.
    @parameter key - optional key function to compute value from each element of N.

    @return - the percentile of the values
    """
    if not N:
        return None
    k = (len(N)-1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return key(N[int(k)])
    d0 = key(N[int(f)]) * (c-k)
    d1 = key(N[int(c)]) * (k-f)
    return d0+d1

@functools.lru_cache(100)
class PercentileAggregator(Aggregator):
    def __init__(self, pct, helpstr=''):
        super().__init__('p%s'%pct, None, helpstr=helpstr)
        self.pct = pct

    def aggregate(self, col, rows):
        return _percentile(sorted(col.getValues(rows)), self.pct/100, key=float)

def quantiles(q, helpstr):
    return [PercentileAggregator(round(100*i/q), helpstr) for i in range(1, q)]

def aggregate_groups(sheet, col, rows, aggr) -> list:
    '''Returns a list, containing the result of the aggregator applied to each row.
    *col* is a column whose values determine each row's rank within a group.
    *rows* is a list of visidata rows.
    *aggr* is an Aggregator object.
    Rows are grouped by their key columns. Null key column cells are considered equal,
    so nulls are grouped together. Cells with exceptions do not group together.
    Each exception cell is grouped by itself, with only one row in the group.
    '''
    def _key_progress(prog):
        def identity(val):
            prog.addProgress(1)
            return val
        return identity

    with Progress(gerund='ranking', total=4*sheet.nRows) as prog:
        p = _key_progress(prog) # increment progress every time p() is called
        # compile row data, for each row a list of tuples: (group_key, rank_key, rownum)
        rowdata = [(sheet.rowkey(r), col.getTypedValue(r), p(rownum)) for rownum, r in enumerate(rows)]
        # sort by row key and column value to prepare for grouping
        try:
            rowdata.sort(key=p)
        except TypeError as e:
            vd.fail(f'elements in a ranking column must be comparable: {e.args[0]}')
        rowvals = []
        #group by row key
        for _, group in itertools.groupby(rowdata, key=lambda v: v[0]):
            # within a group, the rows have already been sorted by col_val
            group = list(group)
            if isinstance(aggr, ListAggregator): # for list aggregators, each row gets its own value
                aggr_vals = aggr.aggregate_list(col, [rows[rownum] for _, _, rownum in group])
                rowvals += [(rownum, v) for (_, _, rownum), v in zip(group, aggr_vals)]
            else:             # for normal aggregators, each row in the group gets the same value
                aggr_val = aggr.aggregate(col, [rows[rownum] for _, _, rownum in group])
                rowvals += [(rownum, aggr_val) for _, _, rownum in group]
            prog.addProgress(len(group))
        # sort by unique rownum, to make rank results match the original row order
        rowvals.sort(key=p)
        rowvals = [ v for rownum, v in rowvals ]
        return rowvals

vd.aggregator('min', min, 'minimum value')
vd.aggregator('max', max, 'maximum value')
vd.aggregator('avg', mean, 'arithmetic mean of values', type=float)
vd.aggregator('mean', mean, 'arithmetic mean of values', type=float)
vd.aggregator('median', statistics.median, 'median of values')
vd.aggregator('mode', statistics.mode, 'mode of values')
vd.aggregator('sum', vsum, 'sum of values')
vd.aggregator('distinct', set, 'distinct values', type=vlen)
vd.aggregator('count', lambda values: sum(1 for v in values), 'number of values', type=int)
vd.aggregator_list('list', 'list of values', type=anytype, listtype=None)
vd.aggregator('stdev', stdev, 'standard deviation of values', type=float)

vd.aggregators['q3'] = quantiles(3, 'tertiles (33/66th pctile)')
vd.aggregators['q4'] = quantiles(4, 'quartiles (25/50/75th pctile)')
vd.aggregators['q5'] = quantiles(5, 'quintiles (20/40/60/80th pctiles)')
vd.aggregators['q10'] = quantiles(10, 'deciles (10/20/30/40/50/60/70/80/90th pctiles)')

# since bb29b6e, a record of every aggregator
# is needed in vd.aggregators
for pct in (10, 20, 25, 30, 33, 40, 50, 60, 67, 70, 75, 80, 90, 95, 99):
    vd.aggregators[f'p{pct}'] = PercentileAggregator(pct, f'{pct}th percentile')

class KeyFindingAggregator(Aggregator):
    '''Return the key of the row that results from applying *aggr_func* to *rows*.
        Return None if *rows* is an empty list.
        *aggr_func* takes a list of (value, row) tuples, one for each row in the column,
        excluding rows where the column holds null and error values.
        *aggr_func* must also take the parameters *default* and *key*, as max() does:
        https://docs.python.org/3/library/functions.html#max'''

    def __init__(self, aggr_func, *args, **kwargs):
        self.aggr_func = aggr_func
        super().__init__(*args, **kwargs)

    def aggregate(self, col, rows):
        if not col.sheet.keyCols:
            vd.error('key aggregator function requires one or more key columns')
            return None
        # convert dicts to lists because functions like max() can't compare dicts
        sortkey = lambda t: (t[0], sorted(t[1].items())) if isinstance(t[1], dict) else t
        row = self.aggr_func(col.getValueRows(rows), default=(None, None), key=sortkey)[1]
        return col.sheet.rowkey(row) if row else None

vd.aggregators['keymin'] = KeyFindingAggregator(min, 'keymin', anytype, helpstr='key of the minimum value')
vd.aggregators['keymax'] = KeyFindingAggregator(max, 'keymax', anytype, helpstr='key of the maximum value')


ColumnsSheet.columns += [
    Column('aggregators',
           getter=lambda c,r:r.aggstr,
           setter=lambda c,r,v:setattr(r, 'aggregators', v),
           help='change the metrics calculated in every Frequency or Pivot derived from the source sheet')
]

@Sheet.api
def addAggregators(sheet, cols, aggrnames):
    'Add each aggregator in list of *aggrnames* to each of *cols*. Ignores names that are not valid.'
    for aggrname in aggrnames:
        aggrs = vd.aggregators.get(aggrname)
        aggrs = aggrs if isinstance(aggrs, list) else [aggrs]
        for c in cols:
            vd.addUndo(setattr, c, 'aggregators', copy(c.aggregators))
            for aggr in aggrs:
                if aggr and aggr not in c.aggregators:
                    c.aggregators += [aggr]


@Column.api
def aggname(col, agg):
    'Consistent formatting of the name of given aggregator for this column.  e.g. "col1_sum"'
    return '%s_%s' % (col.name, agg.name)

@Column.api
def aggregateTotal(col, agg):
    if agg not in col._aggregatedTotals:
        col._aggregatedTotals[agg] = INPROGRESS
        col._aggregateTotalAsync(agg)
    return col._aggregatedTotals[agg]


@Column.api
@asyncthread
def _aggregateTotalAsync(col, agg):
    col._aggregatedTotals[agg] = agg.aggregate(col, col.sheet.rows)


@Column.api
@asyncthread
def memo_aggregate(col, agg_choices, rows):
    'Show aggregated value in status, and add to memory.'
    for agg_choice in agg_choices:
        agg = vd.aggregators.get(agg_choice)
        if not agg: continue
        aggs = agg if isinstance(agg, list) else [agg]
        for agg in aggs:
            aggval = agg.aggregate(col, rows)
            typedval = wrapply(agg.type or col.type, aggval)
            # limit width to limit formatting time when typedval is a long list
            dispval = col.format(typedval, width=1000)
            k = col.name+'_'+agg.name
            vd.status(f'{k}={dispval}')
            vd.memory[k] = typedval


@VisiData.property
def aggregator_choices(vd):
    return [
       AttrDict(key=agg, desc=v[0].helpstr if isinstance(v, list) else v.helpstr)
         for agg, v in vd.aggregators.items()
           if not (agg.startswith('p') and agg[1:].isdigit())  # skip all the percentiles like 'p10', user should use q# instead
    ]


@VisiData.api
def chooseAggregators(vd, prompt = 'choose aggregators: '):
    '''Return a list of aggregator name strings chosen or entered by the user. User-entered names may be invalid.'''
    def _fmt_aggr_summary(match, row, trigger_key):
        formatted_aggrname = match.formatted.get('key', row.key) if match else row.key
        r = ' '*(dispwidth(prompt)-3)
        r += f'[:keystrokes]{trigger_key}[/]  '
        r += formatted_aggrname
        if row.desc:
            r += ' - '
            r += match.formatted.get('desc', row.desc) if match else row.desc
        return r

    r = vd.activeSheet.inputPalette(prompt,
            vd.aggregator_choices,
            value_key='key',
            formatter=_fmt_aggr_summary,
            type='aggregators',
            help=vd.help_aggregators,
            multiple=True)

    aggrs = r.split()
    valid_choices = vd.aggregators.keys()
    for aggr in aggrs:
        vd.usedInputs[aggr] += 1
        if aggr not in valid_choices:
            vd.warning(f'aggregator does not exist: {aggr}')
    return aggrs

@Sheet.api
@asyncthread
def addcol_aggregate(sheet, col, aggrnames):
    for aggrname in aggrnames:
        aggrs = vd.aggregators.get(aggrname)
        aggrs = aggrs if isinstance(aggrs, list) else [aggrs]
        if not aggrs: continue
        for aggr in aggrs:
            rows = aggregate_groups(sheet, col, sheet.rows, aggr)
            if isinstance(aggr, ListAggregator):
                t = aggr.listtype or col.type
            else:
                t = aggr.type or col.type
            c = SettableColumn(name=f'{col.name}_{aggr.name}', type=t)
            sheet.addColumnAtCursor(c)
            c.setValues(sheet.rows, *rows)

Sheet.addCommand('+', 'aggregate-col', 'addAggregators([cursorCol], chooseAggregators())', 'Add aggregator to current column')
Sheet.addCommand('z+', 'memo-aggregate', 'cursorCol.memo_aggregate(chooseAggregators(), selectedRows or rows)', 'memo result of aggregator over values in selected rows for current column')
ColumnsSheet.addCommand('g+', 'aggregate-cols', 'addAggregators(selectedRows or source[0].nonKeyVisibleCols, chooseAggregators())', 'add aggregators to selected source columns')
Sheet.addCommand('', 'addcol-aggregate', 'addcol_aggregate(cursorCol, chooseAggregators(prompt="aggregator for groups: "))', 'add column(s) with aggregator of rows grouped by key columns')

vd.addGlobals(
    ListAggregator=ListAggregator
)

vd.addMenuItems('''
    Column > Add aggregator > aggregate-col
    Column > Add column > aggregate > addcol-aggregate
''')




from visidata import Sheet, VisiData, ItemColumn, vd, AttrDict, Column, setitem, ESC

vd.memory = AttrDict()
vd.contexts += [vd.memory]


@VisiData.api
def memoValue(vd, name, value, dispvalue):
    vd.memory[name] = value
    vd.status(f'memo {name}={dispvalue}')


class MemorySheet(Sheet):
    rowtype = 'memos' # rowdef: keys into vd.memory
    columns = [
        Column('name', getter=lambda c,r: r, setter=lambda c,r,v: setitem(vd.memory, v, vd.memory[r])),
        Column('value', getter=lambda c,r: vd.memory[r], setter=lambda c,r,v: setitem(vd.memory, r, v)),
    ]

    @property
    def rows(self):
        return list(vd.memory.keys())

    @rows.setter
    def rows(self, v):
        pass


@VisiData.lazy_property
def memosSheet(vd):
    return MemorySheet('memos')


@VisiData.api
def inputMemoName(vd, value):
    value_params = dict(prompt="assign value: ", value=value)
    name_params = dict(prompt="to memo name: ")
    r = vd.inputMultiple(memo_name=name_params, memo_value=value_params)
    name = r['memo_name']
    if not name:
        vd.fail('memo name cannot be blank')
    return name


Sheet.addCommand('Alt+Shift+M', 'open-memos', 'vd.push(vd.memosSheet)', 'open the Memory Sheet')
Sheet.addCommand('Alt+m', 'memo-cell', 'vd.memoValue(inputMemoName(cursorDisplay), cursorTypedValue, cursorDisplay)', 'store value in current cell to Memory Sheet')



import io
import pytest

from visidata import Path

class TestVisidataPath:

    def test_withName(self):
        'tests for visidata.Path().with_name'

        file_path = Path('sample_data/sample.tsv')
        url_path = Path('https://visidata.org/hello/sample.tsv')

        assert 'sample_data/b.tsv' == str(file_path.with_name('b.tsv')), '{} should be sample_data/b.tsv'.format(file_path.with_name('b.tsv'))
        assert 'sample_data/a/b.tsv' == str(file_path.with_name('a/b.tsv')), '{} should be sample_data/a/b.tsv'.format(file_path.with_name('a/b.tsv'))

        assert "https://visidata.org/hello/b.tsv" == str(url_path.with_name('b.tsv')), '{} should be https://visidata.org/hello/b.tsv'.format(url_path.with_name('b.tsv'))
        assert "https://visidata.org/hello/a/b.tsv" == str(url_path.with_name('a/b.tsv')), '{} should be https://visidata.org/hello/a/b.tsv'.format(url_path.with_name('a/b.tsv'))

        assert Path('foo.a.b').base_stem == 'foo.a'
        assert Path('foo.a.b').ext == 'b'
        assert Path('foo').ext == ''
        assert Path('foo').base_stem == 'foo'
        assert Path('foo.').ext == ''
        # assert Path('foo.').base_stem == 'foo.' # only 'foo' since python 3.14
        assert Path('.foo').ext == ''
        assert Path('.foo').base_stem == '.foo'


    def test_opentwice(self):
        'fresh iterator for each open'
        p = Path('test', fptext=io.StringIO('<html>'))
        a = next(p.open())
        b = next(p.open())
        assert a == b



from copy import copy
from visidata import VisiData, vd, Sheet, options, Column, Progress, setitem, ColumnAttr, vlen, RowColorizer, Path

vd.option('xml_parser_huge_tree', True, 'allow very deep trees and very long text content')


@VisiData.api
def open_xml(vd, p):
    return XmlSheet(p.base_stem, source=p)

VisiData.open_svg = VisiData.open_xml

def unns(k):
    'de-namespace key k'
    if '}' in k:
        return k[k.find('}')+1:]
    return k


def AttribColumn(name, k, **kwargs):
    return Column(name, getter=lambda c,r,k=k: r.attrib.get(k),
                        setter=lambda c,r,v,k=k: setitem(r.attrib, k, v), **kwargs)


# source is Path or xml.Element; root is xml.Element
class XmlSheet(Sheet):
    rowtype = 'elements'   # rowdef: lxml.xml.Element

    columns = [
        ColumnAttr('sourceline', type=int, width=0),
        ColumnAttr('prefix', width=0),
        ColumnAttr('nstag', 'tag', width=0),
        Column('path', width=0, getter=lambda c,r: c.sheet.root.getpath(r)),
        Column('tag', getter=lambda c,r: unns(r.tag)),
        Column('children', type=vlen, getter=lambda c,r: r.getchildren()),
        ColumnAttr('text'),
        ColumnAttr('tail', width=0),
    ]
    colorizers = [
            RowColorizer(8, None, lambda s,c,r,v: 'green' if r is s.source else None)
    ]

    def showColumnsBasedOnRow(self, row):
        for c in self.columns:
            nstag = getattr(c, 'nstag', '')
            if nstag:
                c.hide(nstag not in row.attrib)

    def iterload(self):
        if isinstance(self.source, Path):
            vd.importExternal('lxml')
            from lxml import etree, objectify
            p = etree.XMLParser(**self.options.getall('xml_parser_'))
            with self.open_text_source() as fp:
                self.root = etree.parse(fp, parser=p)
            objectify.deannotate(self.root, cleanup_namespaces=True)
        else: #        elif isinstance(self.source, XmlElement):
            self.root = self.source

        self.attribcols = {}
        self.columns = []
        for c in XmlSheet.columns:
            self.addColumn(copy(c))

        if getattr(self.root, 'iterancestors', None):
            for elem in Progress(list(self.root.iterancestors())[::-1]):
                yield elem

        for elem in self.root.iter():
            yield elem

    def openRow(self, row):
        return XmlSheet("%s_%s" % (unns(row.tag), row.attrib.get("id")), source=row)

    def addRow(self, elem):
        super().addRow(elem)
        for k in elem.attrib:
            if k not in self.attribcols:
                c = AttribColumn(unns(k), k)
                self.addColumn(c)
                self.attribcols[k] = c
                c.nstag = k


@VisiData.api
def save_xml(vd, p, vs):
    isinstance(vs, XmlSheet) or vd.fail('must save xml from XmlSheet')
    vs.root.write(str(p), encoding=vs.options.save_encoding, standalone=False, pretty_print=True)

XmlSheet.options.save_encoding = 'utf-8'  #2520

XmlSheet.addCommand('za', 'addcol-xmlattr', 'attr=input("add attribute: "); addColumnAtCursor(AttribColumn(attr, attr))', 'add column for xml attribute')
XmlSheet.addCommand('v', 'visibility', 'showColumnsBasedOnRow(cursorRow)', 'show only columns in current row attributes')



from copy import copy

from visidata import Sheet, vd, asyncsingle

@Sheet.api
def dup_search(sheet, cols='cursorCol'):
    vs = copy(sheet)
    vs.name += "_search"
    vs.rows = sheet.rows
    vs.source = sheet
    vs.search = ''

    @asyncsingle
    def live_search_async(val, status=False):
        if not val:
            vs.rows = vs.source.rows
        else:
            vs.rows = []
            for i in vd.searchRegex(vs.source, regex=val, columns=cols, printStatus=status):
                vs.addRow(vs.source.rows[i])

    def live_search(val):
        vs.draw(vs._scr)
        vd.drawRightStatus(vs._scr, vs)
        val = val.rstrip('\n')
        if val == vs.search:
            return
        vs.search = val
        live_search_async(val, sheet=vs, status=False)

    vd.input("search regex: ", updater=live_search)
    vd.push(vs)
    vs.name = vs.source.name+'_'+vs.search


Sheet.addCommand('^[s', 'dup-search', 'dup_search("cursorCol")', 'search for regex forwards in current column, creating duplicate sheet with matching rows live')
Sheet.addCommand('g^[s', 'dup-search-cols', 'dup_search("visibleCols")', 'search for regex forwards in all columns, creating duplicate sheet with matching rows live')



from copy import copy

from visidata import vd, VisiData, asyncthread
from visidata import Sheet, RowColorizer, CellColorizer, Column, BaseSheet, Progress

vd.theme_option('color_add_pending', 'green', 'color for rows pending add')
vd.theme_option('color_change_pending', 'reverse yellow', 'color for cells pending modification')
vd.theme_option('color_delete_pending', 'red', 'color for rows pending delete')
vd.option('overwrite', 'c', 'overwrite existing files {y=yes|c=confirm|n=no}')

vd.optalias('readonly', 'overwrite', 'n')
vd.optalias('ro', 'overwrite', 'n')
vd.optalias('y', 'overwrite', 'y')


@VisiData.api
def couldOverwrite(vd) -> bool:
    'Return True if overwrite might be allowed.'
    return vd.options.overwrite.startswith(('y','c'))


@VisiData.api
def confirmOverwrite(vd, path, msg:str=''):
    'Fail if file exists and overwrite not allowed.'
    if path is None or path.exists():
        msg = msg or f'{path.given} exists. overwrite? '
        ow = vd.options.overwrite
        if ow.startswith('c'):  # confirm
            vd.confirm(msg)
        elif ow.startswith('y'):  # yes/always
            pass
        else: #1805  empty/no/never/readonly
            vd.fail('overwrite disabled')
    return True

# deferred cached
@Sheet.lazy_property
def _deferredAdds(sheet):
    return dict() # [s.rowid(row)] -> row

@Sheet.lazy_property
def _deferredMods(sheet):
    return dict() # [s.rowid(row)] -> (row, { [col] -> val })

@Sheet.lazy_property
def _deferredDels(sheet):
    return dict() # [s.rowid(row)] -> row

Sheet.colorizers += [
        RowColorizer(9, 'color_add_pending', lambda s,c,r,v: s.rowid(r) in s._deferredAdds),
        CellColorizer(8, 'color_change_pending', lambda s,c,r,v: c and (r is not None) and s.isChanged(c, r)),
        RowColorizer(9, 'color_delete_pending', lambda s,c,r,v: s.isDeleted(r)),
        ]

@Sheet.api
def preloadHook(sheet):
    BaseSheet.preloadHook(sheet)
    sheet._deferredAdds.clear()
    sheet._deferredMods.clear()
    sheet._deferredDels.clear()

@Sheet.api
def rowAdded(self, row):
    'Mark row as a deferred add-row'
    self._deferredAdds[self.rowid(row)] = row
    def _undoRowAdded(sheet, row):
        if sheet.rowid(row) not in sheet._deferredAdds:
            vd.warning('cannot undo row addition after a commit')
            return
        del sheet._deferredAdds[sheet.rowid(row)]
    vd.addUndo(_undoRowAdded, self, row)

@Column.api
def cellChanged(col, row, val):
    'Mark cell at row for col as a deferred edit-cell'
    oldval = col.getValue(row)
    if oldval != val:
        rowid = col.sheet.rowid(row)
        if rowid not in col.sheet._deferredMods:
            rowmods = {}
            col.sheet._deferredMods[rowid] = (row, rowmods)
        else:
            _, rowmods = col.sheet._deferredMods[rowid]
        rowmods[col] = val

        def _undoCellChanged(col, row, oldval):
            if oldval == col.getSourceValue(row):
                # if we have reached the original value, remove from defermods entirely
                if col.sheet.rowid(row) not in col.sheet._deferredMods:
                    vd.warning('cannot undo cell change after a commit')
                    return
                del col.sheet._deferredMods[col.sheet.rowid(row)]
            else:
                # otherwise, update deferredMods with previous value
                _, rowmods = col.sheet._deferredMods[col.sheet.rowid(row)]
                rowmods[col] = oldval

        vd.addUndo(_undoCellChanged, col, row, oldval)

@Sheet.api
def rowDeleted(self, row):
    'Mark row as a deferred delete-row'
    self._deferredDels[self.rowid(row)] = row
    self.addUndoSelection()
    self.unselectRow(row)
    def _undoRowDeleted(sheet, row):
        if sheet.rowid(row) not in sheet._deferredDels:
            vd.warning('cannot undo row deletion after a commit')
            return
        del sheet._deferredDels[sheet.rowid(row)]
    vd.addUndo(_undoRowDeleted, self, row)


@Sheet.api
@asyncthread
def addRows(sheet, rows, index=None, undo=True):
    'Add *rows* after row at *index*.'
    addedRows = {}
    if index is None: index=len(sheet.rows)
    for i, row in enumerate(Progress(rows, gerund='adding')):
        addedRows[sheet.rowid(row)] = row
        sheet.addRow(row, index=index+i+1)

        if sheet.defer:
            sheet.rowAdded(row)
    sheet.setModified()

    @asyncthread
    def _removeRows():
        sheet.deleteBy(lambda r,sheet=sheet,addedRows=addedRows: sheet.rowid(r) in addedRows, commit=True, undo=False)

    if undo:
        vd.addUndo(_removeRows)


@Sheet.api
def deleteBy(sheet, func, commit=False, undo=True):
    '''Delete rows on sheet for which ``func(row)`` returns true.  Return number of rows deleted.
    If sheet.defer is set and *commit* is True, remove rows immediately without deferring.
    If undo is set to True, add an undo for deletion.'''
    oldrows = copy(sheet.rows)
    oldidx = sheet.cursorRowIndex
    ndeleted = 0

    newCursorRow = None   # row to re-place cursor after
    # if commit is True, commit to delete, even if defer is True
    if sheet.defer and not commit:
        ndeleted = 0
        for r in sheet.gatherBy(func, 'deleting'):
            sheet.rowDeleted(r)
            ndeleted += 1
        return ndeleted

    # find next non-deleted row to go to once delete has finished
    while oldidx < len(oldrows):
        if not func(oldrows[oldidx]):
            newCursorRow = sheet.rows[oldidx]
            break
        oldidx += 1

    sheet.rows.clear() # must delete from the existing rows object
    for r in Progress(oldrows, 'deleting'):
        if not func(r):
            sheet.rows.append(r)
            if r is newCursorRow:
                sheet.cursorRowIndex = len(sheet.rows)-1
        else:
            try:
                sheet.commitDeleteRow(r)
                ndeleted += 1
            except Exception as e:
                vd.exceptionCaught(e)

    if undo:
        vd.addUndo(setattr, sheet, 'rows', oldrows)
        sheet.setModified()

    if ndeleted:
        vd.status('deleted %s %s' % (ndeleted, sheet.rowtype))

    return ndeleted


@Sheet.api
def isDeleted(self, row):
    'Return True if *row* has been deferred for deletion.'
    return self.rowid(row) in self._deferredDels

@Sheet.api
def isChanged(self, col, row):
    'Return True if cell at *row* for *col* has been deferred for modification.'
    try:
        row, rowmods = self._deferredMods[self.rowid(row)]
        newval = rowmods[col]
        curval = col.getSourceValue(row)
        return (newval is None and curval is not None) or (curval is None and newval is not None) or (col.type(newval) != col.type(curval))
    except KeyError:
        return False
    except Exception:
        return False

@Column.api
def getSourceValue(col, row):
    'For deferred sheets, return value for *row* in this *col* as it would be in the source, without any deferred modifications applied.'
    return Column.calcValue(col, row)


@Sheet.api
def commitAdds(self):
    'Return the number of rows that have been marked for deferred add-row. Clear the marking.'
    nadded = 0
    nerrors = 0
    for row in self._deferredAdds.values():
        try:
            self.commitAddRow(row)
            nadded += 1
        except Exception as e:
            vd.exceptionCaught(e)
            nerrors += 1

    if nadded or nerrors:
        vd.status(f'added {nadded} {self.rowtype} ({nerrors} errors)')

    self._deferredAdds.clear()
    return nadded


@Sheet.api
def commitMods(sheet):
    'Commit all deferred modifications (that are not from rows added or deleted in this commit.  Return number of cells changed.'
    _, deferredmods, _ = sheet.getDeferredChanges()

    nmods = 0
    for row, rowmods in deferredmods.values():
        for col, val in rowmods.items():
            try:
                col.putValue(row, val)
                nmods += 1
            except Exception as e:
                vd.exceptionCaught(e)

    sheet._deferredMods.clear()
    return nmods

@Sheet.api
def commitDeletes(self):
    'Return the number of rows that have been marked for deletion. Delete the rows. Clear the marking.'
    ndeleted = self.deleteBy(self.isDeleted, commit=True, undo=False)

    if ndeleted:
        vd.status('deleted %s %s' % (ndeleted, self.rowtype))
    return ndeleted


@Sheet.api
def commitAddRow(self, row):
    'To commit an added row.  Override per sheet type.'


@Sheet.api
def commitDeleteRow(self, row):
    'To commit a deleted row.  Override per sheet type.'


@asyncthread
@Sheet.api
def putChanges(sheet):
    'Commit changes to ``sheet.source``. May overwrite source completely without confirmation.  Overridable.'
    sheet.commitAdds()
    sheet.commitMods()
    sheet.commitDeletes()

    # clear after save, to ensure cstr (in commit()) is aware of deletes
    sheet._deferredDels.clear()

@Sheet.api
def getDeferredChanges(sheet):
    '''Return changes made to deferred sheets that have not been committed, as a tuple (added_rows, modified_rows, deleted_rows).  *modified_rows* does not include any *added_rows* or *deleted_rows*.

        - *added_rows*: { rowid:row, ... }
        - *modified_rows*: { rowid: (row, { col:val, ... }), ... }
        - *deleted_rows*: { rowid: row }

    *rowid* is from ``Sheet.rowid(row)``. *col* is an actual Column object.
    '''

    # only report mods if they aren't adds or deletes
    mods = {} # [rowid] -> (row, dict(col:val))
    for row, rowmods in sheet._deferredMods.values():
        rowid = sheet.rowid(row)
        if rowid not in sheet._deferredAdds and rowid not in sheet._deferredDels:
            mods[rowid] = (row, {col:val for col, val in rowmods.items() if sheet.isChanged(col, row)})

    return sheet._deferredAdds, mods, sheet._deferredDels

@Sheet.api
def changestr(self, adds, mods, deletes):
    'Return a str for status that outlines how many deferred changes are going to be committed.'
    cstr = ''
    if adds:
        cstr += 'add %d %s' % (len(adds), self.rowtype)

    if mods:
        if cstr: cstr += ' and '
        cstr += 'change %d values' % sum(len(rowmods) for row, rowmods in mods.values())

    if deletes:
        if cstr: cstr += ' and '
        cstr += 'delete %d %s' % (len(deletes), self.rowtype)

    return cstr

@Sheet.api
def commit(sheet, *rows):
    'Commit all deferred changes on this sheet to original ``sheet.source``.'
    if not sheet.defer:
        vd.fail('commit-sheet is not enabled for this sheet type')

    adds, mods, deletes = sheet.getDeferredChanges()
    cstr = sheet.changestr(adds, mods, deletes)

    vd.confirmOverwrite(sheet.rootSheet().source, 'really ' + cstr + '? ')

    sheet.putChanges()
    sheet.hasBeenModified = False


@Sheet.api
def new_rows(sheet, n):
    return [sheet.newRow() for i in range(n)]

Sheet.addCommand('a', 'add-row', 'addRows([newRow()], index=cursorRowIndex); cursorDown(1)', 'append a blank row')
Sheet.addCommand('ga', 'add-rows', 'n=int(input("add # rows: ", value=1)); addRows(new_rows(n), index=cursorRowIndex); cursorDown(1)', 'append N blank rows')
Sheet.addCommand('za', 'addcol-new', 'addColumnAtCursor(SettableColumn(input("column name: ")))', 'append an empty column')
Sheet.addCommand('gza', 'addcol-bulk', 'addColumnAtCursor(*(SettableColumn() for c in range(int(input("add columns: ")))))', 'append N empty columns')

Sheet.addCommand('z^S', 'commit-sheet', 'commit()', 'commit changes back to source.  not undoable!')

vd.addMenuItems('''
    File > Save > changes to source > commit-sheet
    Row > Add > one row
    Row > Add > multiple rows
    Column > Add column > empty > one column > addcol-new
    Column > Add column > empty > multiple columns > addcol-bulk
''')
